from .dev_skip_test import dev_skip_test, istest, nottest

d = {}

@istest
def test_passes():
    d['istest'] = True
    assert True

@nottest
def test_fails():
    d['nottest'] = True
    assert False

def test_dev_skip_test():
    assert dev_skip_test is istest or dev_skip_test is nottest
    assert d.get('istest', False) is True
    assert d.get('nottest', False) is False



from .itertoolz import frequencies, getter, pluck
import itertools

__all__ = ('countby', 'partitionby')

def countby(key, seq):
    """ Count elements of a collection by a key function

    >>> countby(len, ['cat', 'mouse', 'dog'])
    {3: 2, 5: 1}

    >>> def iseven(x): return x % 2 == 0
    >>> countby(iseven, [1, 2, 3])  # doctest:+SKIP
    {True: 1, False: 2}

    See Also:
        groupby
    """
    if not callable(key):
        key = getter(key)
    return frequencies(map(key, seq))

def partitionby(func, seq):
    """ Partition a sequence according to a function

    Partition `s` into a sequence of lists such that, when traversing
    `s`, every time the output of `func` changes a new list is started
    and that and subsequent items are collected into that list.

    >>> is_space = lambda c: c == " "
    >>> list(partitionby(is_space, "I have space"))
    [('I',), (' ',), ('h', 'a', 'v', 'e'), (' ',), ('s', 'p', 'a', 'c', 'e')]

    >>> is_large = lambda x: x > 10
    >>> list(partitionby(is_large, [1, 2, 1, 99, 88, 33, 99, -1, 5]))
    [(1, 2, 1), (99, 88, 33, 99), (-1, 5)]

    See also:
        partition
        groupby
        itertools.groupby
    """
    return map(tuple, pluck(1, itertools.groupby(seq, key=func)))



from humpy_toolz.functoolz import curry
# Unary operators and special cases - not curried (from IGNORE set in operator.py)
from operator import (
	__abs__ as __abs__, __index__ as __index__, __inv__ as __inv__, __invert__ as __invert__, __neg__ as __neg__, __not__ as __not__,
	__pos__ as __pos__, abs as abs, attrgetter as attrgetter, index as index, inv as inv, invert as invert, itemgetter as itemgetter,
	neg as neg, not_ as not_, pos as pos, truth as truth)
import operator

__all__ = [
	# Unary operators and special cases (not curried)
	"__abs__",
	"abs",
	"__index__",
	"index",
	"__inv__",
	"inv",
	"__invert__",
	"invert",
	"__neg__",
	"neg",
	"__not__",
	"not_",
	"__pos__",
	"pos",
	"truth",
	"attrgetter",
	"itemgetter",
	# Binary and n-ary operators (curried)
	"__add__",
	"add",
	"__and__",
	"and_",
	"__call__",
	"call",
	"__concat__",
	"concat",
	"__contains__",
	"contains",
	"countOf",
	"__delitem__",
	"delitem",
	"__eq__",
	"eq",
	"__floordiv__",
	"floordiv",
	"__ge__",
	"ge",
	"__getitem__",
	"getitem",
	"__gt__",
	"gt",
	"__iadd__",
	"iadd",
	"__iand__",
	"iand",
	"__iconcat__",
	"iconcat",
	"__ifloordiv__",
	"ifloordiv",
	"__ilshift__",
	"ilshift",
	"__imatmul__",
	"imatmul",
	"__imod__",
	"imod",
	"__imul__",
	"imul",
	"indexOf",
	"__ior__",
	"ior",
	"__ipow__",
	"ipow",
	"__irshift__",
	"irshift",
	"is_",
	"is_not",
	"__isub__",
	"isub",
	"__itruediv__",
	"itruediv",
	"__ixor__",
	"ixor",
	"__le__",
	"le",
	"length_hint",
	"__lshift__",
	"lshift",
	"__lt__",
	"lt",
	"__matmul__",
	"matmul",
	"methodcaller",
	"__mod__",
	"mod",
	"__mul__",
	"mul",
	"__ne__",
	"ne",
	"__or__",
	"or_",
	"__pow__",
	"pow",
	"__rshift__",
	"rshift",
	"__setitem__",
	"setitem",
	"__sub__",
	"sub",
	"__truediv__",
	"truediv",
	"__xor__",
	"xor",
]

# Binary and n-ary operators - curried
# Define non-dunder versions (canonical), then alias dunder versions

# Arithmetic operators
add = curry(operator.add)
__add__ = add

sub = curry(operator.sub)
__sub__ = sub

mul = curry(operator.mul)
__mul__ = mul

truediv = curry(operator.truediv)
__truediv__ = truediv

floordiv = curry(operator.floordiv)
__floordiv__ = floordiv

mod = curry(operator.mod)
__mod__ = mod

pow = curry(operator.pow)
__pow__ = pow

matmul = curry(operator.matmul)
__matmul__ = matmul

# Bitwise operators
and_ = curry(operator.and_)
__and__ = and_

or_ = curry(operator.or_)
__or__ = or_

xor = curry(operator.xor)
__xor__ = xor

lshift = curry(operator.lshift)
__lshift__ = lshift

rshift = curry(operator.rshift)
__rshift__ = rshift

# Comparison operators
eq = curry(operator.eq)
__eq__ = eq

ne = curry(operator.ne)
__ne__ = ne

lt = curry(operator.lt)
__lt__ = lt

le = curry(operator.le)
__le__ = le

gt = curry(operator.gt)
__gt__ = gt

ge = curry(operator.ge)
__ge__ = ge

# In-place operators
iadd = curry(operator.iadd)
__iadd__ = iadd

isub = curry(operator.isub)
__isub__ = isub

imul = curry(operator.imul)
__imul__ = imul

itruediv = curry(operator.itruediv)
__itruediv__ = itruediv

ifloordiv = curry(operator.ifloordiv)
__ifloordiv__ = ifloordiv

imod = curry(operator.imod)
__imod__ = imod

ipow = curry(operator.ipow)
__ipow__ = ipow

imatmul = curry(operator.imatmul)
__imatmul__ = imatmul

iand = curry(operator.iand)
__iand__ = iand

ior = curry(operator.ior)
__ior__ = ior

ixor = curry(operator.ixor)
__ixor__ = ixor

ilshift = curry(operator.ilshift)
__ilshift__ = ilshift

irshift = curry(operator.irshift)
__irshift__ = irshift

# Sequence/container operators
concat = curry(operator.concat)
__concat__ = concat

iconcat = curry(operator.iconcat)
__iconcat__ = iconcat

contains = curry(operator.contains)
__contains__ = contains

getitem = curry(operator.getitem)
__getitem__ = getitem

setitem = curry(operator.setitem)
__setitem__ = setitem

delitem = curry(operator.delitem)
__delitem__ = delitem

# Other binary operators
is_ = curry(operator.is_)
is_not = curry(operator.is_not)

call = curry(operator.call)
__call__ = call

# Utility functions
countOf = curry(operator.countOf)
indexOf = curry(operator.indexOf)
length_hint = curry(operator.length_hint)
methodcaller = curry(operator.methodcaller)



# See #166: https://github.com/pytoolz/toolz/issues/166
# See #173: https://github.com/pytoolz/toolz/pull/173
from collections.abc import Callable, Hashable, Iterable, Iterator
from typing import overload, override

class EqualityHashKey:
	def __init__[T](
		self, key: Callable[[T], Hashable] | int | None, item: T
	) -> None: ...
	@override
	def __hash__(self) -> int: ...
	@override
	def __eq__(self, other: object) -> bool: ...
	@override
	def __ne__(self, other: object) -> bool: ...

# See issue #293: https://github.com/pytoolz/toolz/issues/239
@overload
def unzip(seq: Iterable[tuple[()]]) -> tuple[()]: ...
@overload
def unzip[T1](
	seq: Iterable[tuple[T1]],
) -> tuple[Iterator[T1]]: ...
@overload
def unzip[T1, T2](
	seq: Iterable[tuple[T1, T2]],
) -> tuple[Iterator[T1], Iterator[T2]]: ...
@overload
def unzip[T1, T2, T3](
	seq: Iterable[tuple[T1, T2, T3]],
) -> tuple[
	Iterator[T1],
	Iterator[T2],
	Iterator[T3],
]: ...
@overload
def unzip[T1, T2, T3, T4](
	seq: Iterable[tuple[T1, T2, T3, T4]],
) -> tuple[
	Iterator[T1],
	Iterator[T2],
	Iterator[T3],
	Iterator[T4],
]: ...
@overload
def unzip[T](
	seq: Iterable[tuple[T, ...]],
) -> tuple[Iterator[T], ...]: ...

# Implementation signature
def unzip[T](
	seq: Iterable[tuple[T, ...]],
) -> tuple[Iterator[T], ...]:
	...



"""Package configuration and defensive programming utilities for Python projects."""
from importlib.util import find_spec
from pathlib import Path
from tomllib import loads as tomllib_loads
from typing import TYPE_CHECKING
import dataclasses

if TYPE_CHECKING:
	from importlib.machinery import ModuleSpec

def getIdentifierPackagePACKAGING(identifierPackageFALLBACK: str) -> str:
	"""Get package name from pyproject.toml or fallback to provided value."""
	try:
		return tomllib_loads(Path('pyproject.toml').read_text(encoding='utf-8'))['project']['name']
	except Exception:  # noqa: BLE001
		return identifierPackageFALLBACK

def getPathPackageINSTALLING(identifierPackage: str) -> Path:
	"""Return the root directory of the installed package."""
	try:
		moduleSpecification: ModuleSpec | None = find_spec(identifierPackage)
		if moduleSpecification and moduleSpecification.origin:
			pathFilename: Path = Path(moduleSpecification.origin)
			return pathFilename.parent if pathFilename.is_file() else pathFilename
	except ModuleNotFoundError:
		pass
	return Path.cwd()

@dataclasses.dataclass
class PackageSettings:
	"""Configuration container for Python package metadata and runtime settings.

	This `class` provides a simple way to store and access basic information about a Python package, It will automatically resolve
	package identifiers and installation paths if they are not passed to the `class` constructor. Python `dataclasses` are easy to
	subtype and extend.

	Attributes
	----------
	identifierPackageFALLBACK : str = ''
		Fallback package identifier used only during initialization when automatic discovery fails.
	identifierPackage : str = ''
		Canonical name of the package. Automatically extracted from "pyproject.toml".
	pathPackage : Path = getPathPackageINSTALLING(identifierPackage)
		Absolute path to the installed package directory. Automatically resolved from `identifierPackage` if not provided.
	fileExtension : str = '.py'
		Default file extension.

	Examples
	--------
	Automatic package discovery from development environment:

	```python
	settings = PackageSettings(identifierPackageFALLBACK='cobraPy')
	# Automatically discovers package name from pyproject.toml
	# Resolves installation path from package identifier
	```

	Explicit configuration for specific deployment:

	```python
	settings = PackageSettings(
		identifierPackage='cobraPy',
		pathPackage=Path('/opt/tenEx/packages/cobraPy'),
		fileExtension='.pyx'
	)
	```

	"""

	identifierPackageFALLBACK: dataclasses.InitVar[str] = ''
	"""Fallback package identifier used during initialization only."""
	pathPackage: Path = dataclasses.field(default_factory=Path, metadata={'evaluateWhen': 'installing'})
	"""Absolute path to the installed package."""
	identifierPackage: str = dataclasses.field(default='', metadata={'evaluateWhen': 'packaging'})
	"""Name of this package."""
	fileExtension: str = dataclasses.field(default='.py', metadata={'evaluateWhen': 'installing'})
	"""Default file extension for files."""

	def __post_init__(self, identifierPackageFALLBACK: str) -> None:
		"""Initialize computed fields after dataclass initialization."""
		if not self.identifierPackage and identifierPackageFALLBACK:
			self.identifierPackage = getIdentifierPackagePACKAGING(identifierPackageFALLBACK)
		if self.pathPackage == Path() and self.identifierPackage:
			self.pathPackage = getPathPackageINSTALLING(self.identifierPackage)

def raiseIfNone[TypeSansNone](expression: TypeSansNone | None, errorMessage: str | None = None) -> TypeSansNone:
	"""Convert the `expression` return annotation from '`cerPytainty | None`' to '`cerPytainty`' because `expression` cannot be `None`; `raise` an `Exception` if you're wrong.

	The Python interpreter evaluates `expression` to a value: think of a function call or an attribute access. You can use
	`raiseIfNone` for fail early defensive programming. I use it, however, to cure type-checker-nihilism: that's when "or `None`"
	return types cause your type checker to repeatedly say, "You can't do that because the value might be `None`."

	Parameters
	----------
	expression : TypeSansNone | None
		Python code with a return type that is a `union` of `None` and `TypeSansNone`, which is a stand-in for one or more other types.
	errorMessage : str | None = 'A function unexpectedly returned `None`. Hint: look at the traceback immediately before `raiseIfNone`.'
		Custom error message for the `ValueError` `Exception` if `expression` is `None`.

	Returns
	-------
	contentment : TypeSansNone
		The value returned by `expression`, but guaranteed to not be `None`.

	Raises
	------
	ValueError
		If the value returned by `expression` is `None`.

	Examples
	--------
	Basic usage with attribute access:
	```python
	annotation = raiseIfNone(ast_arg.annotation)
	# Raises ValueError if ast_arg.annotation is None
	```

	Function return value validation:
	```python
	def findFirstMatch(listItems: list[str], pattern: str) -> str | None:
		for item in listItems:
			if pattern in item:
				return item
		return None

	listFiles = ['document.txt', 'image.png', 'data.csv']
	filename = raiseIfNone(findFirstMatch(listFiles, '.txt'))
	# Returns 'document.txt' when match exists
	```

	Dictionary value retrieval with custom message:
	```python
	configurationMapping = {'host': 'localhost', 'port': 8080}
	host = raiseIfNone(configurationMapping.get('host'),
					"Configuration must include 'host' setting")
	# Returns 'localhost' when key exists

	# This would raise ValueError with custom message:
	# database = raiseIfNone(configurationMapping.get('database'),
	#                      "Configuration must include 'database' setting")
	```

	Thanks
	------
	sobolevn, https://github.com/sobolevn, for the seed of this function. https://github.com/python/typing/discussions/1997#discussioncomment-13108399

	"""
	if expression is None:
		message: str = errorMessage or 'A function unexpectedly returned `None`. Hint: look at the traceback immediately before `raiseIfNone`.'
		raise ValueError(message)
	return expression



from collections.abc import Callable, Iterable
from typing import Any

__all__ = ("consume", "include_dirs", "no_default", "raises")

def raises(err: type[Exception], lamda: Callable[[], Any]) -> bool: ...

no_default: Any = "__no__default__"

def include_dirs() -> tuple[str, str]: ...
def consume(seq: Iterable[Any]) -> None: ...



from types import ModuleType
import humpy_toolz

def test_tlz() -> None:
    import humpy_tlz
    humpy_tlz.curry
    humpy_tlz.functoolz.curry
    assert humpy_tlz.__package__ == 'humpy_tlz'
    assert humpy_tlz.__name__ == 'humpy_tlz'
    import humpy_tlz.curried
    assert humpy_tlz.curried.__package__ == 'humpy_tlz.curried'
    assert humpy_tlz.curried.__name__ == 'humpy_tlz.curried'
    humpy_tlz.curried.curry
    import humpy_tlz.curried.operator
    assert humpy_tlz.curried.operator.__package__ in (None, 'humpy_tlz.curried')
    assert humpy_tlz.curried.operator.__name__ == 'humpy_tlz.curried.operator'
    assert humpy_tlz.functoolz.__name__ == 'humpy_tlz.functoolz'
    m1: ModuleType = humpy_tlz.functoolz
    import humpy_tlz.functoolz as m2
    assert m1 is m2
    import humpy_tlz.sandbox
    try:
        import tlzthisisabadname.curried
        1 / 0
    except ImportError:
        pass
    try:
        import tlz.curry
        1 / 0
    except ImportError:
        pass
    try:
        import tlz.badsubmodulename
        1 / 0
    except ImportError:
        pass
    assert humpy_toolz.__package__ == 'humpy_toolz'
    assert humpy_toolz.curried.__package__ == 'humpy_toolz.curried'
    assert humpy_toolz.functoolz.__name__ == 'humpy_toolz.functoolz'
    try:
        import humpy_cytoolz
        assert humpy_cytoolz.__package__ == 'humpy_cytoolz'
        assert humpy_cytoolz.curried.__package__ == 'humpy_cytoolz.curried'
        assert humpy_cytoolz.functoolz.__name__ == 'humpy_cytoolz.functoolz'
    except ImportError:
        pass
    if hasattr(humpy_tlz, '__file__'):
        assert humpy_tlz.__file__ == humpy_toolz.__file__
    if hasattr(humpy_tlz.functoolz, '__file__'):
        assert humpy_tlz.functoolz.__file__ == humpy_toolz.functoolz.__file__
    assert humpy_tlz.pipe is humpy_toolz.pipe
    assert 'humpy_tlz' in humpy_tlz.__doc__
    assert humpy_tlz.curried.__doc__ is not None



"""Express semantic intent through named constants and ANSI color codes.

(AI generated docstring)

You can use this module to replace numeric literals with semantically meaningful identifiers
and apply colored terminal output using high-contrast ANSI escape sequences [1]. The module
provides semantic constants for directional iteration, boundary inclusion, indexing
convention adjustments, and error state signaling, along with terminal color formatting
utilities.

The semantic constants make code intent explicit by replacing ambiguous numeric literals like
`-1`, `1`, and `31212012` with descriptive identifiers such as `decreasing`, `inclusive`,
`zeroIndexed`, and `errorL33T`. The ANSI color utilities provide escape sequences for
high-contrast color combinations selected through visibility research [2], enabling clear
terminal output during debugging and logging operations.

Contents
--------
Semantic Constants
	decreasing
		Express descending iteration or a reverse direction.
	errorL33T
		Signal an error state with a visually distinctive numeric value.
	inclusive
		Express inclusion or exclusion of a boundary value.
	zeroIndexed
		Express that the adjustment to a value is due to zero-based indexing.

ANSI Terminal Color
	ansiColorReset
		Reset terminal text color and background to default settings.
	AnsiColors
		Provide high-contrast ANSI color combinations for terminal output.

References
----------
[1] ANSI escape code - Wikipedia
	https://en.wikipedia.org/wiki/ANSI_escape_code
[2] Color contrast - Wikipedia
	https://en.wikipedia.org/wiki/Contrast_(vision)#Color_contrast

"""
from typing import NamedTuple

#======== Semantic replacements for ambiguous values. =======

decreasing: int = -1
"""Express descending iteration or a reverse direction.

(AI generated docstring)

The identifier `decreasing` holds the value `-1` and serves as a semantic replacement for
numeric literals in contexts where direction or ordering matters. You can use `decreasing`
as an addend to adjust boundary values, as a multiplicand to reverse sign or direction,
or in both roles simultaneously to express complex transformations.

You can use `decreasing` wherever `-1` would appear but where the meaning "descending order"
or "reverse direction" is more important than the specific numeric value. Using `decreasing`
makes the code's intent explicit and communicates the semantic purpose to readers who might
not immediately recognize `-1` as a directional indicator.

Common contexts include: reverse iteration through sequences, computing predecessors or
backward offsets, negating dimensions or indices, and constructing `range` objects that
count downward.

Examples
--------
You can use `decreasing` as an addend to compute a loop boundary:

>>> for countDown in range(dimension - 2 + decreasing, decreasing, decreasing):
...     processValue(countDown)

In this example, `dimension - 2 + decreasing` computes the starting index (equivalent to
`dimension - 3`), the second `decreasing` stops iteration at `-1` (to include `0`), and
the third `decreasing` specifies the step direction (counting down by `1`).

You can use `decreasing` as a multiplicand to reverse sign:

>>> negatedOffset = baseOffset * decreasing

You can use `decreasing` as both multiplicand and addend in a single expression:

>>> adjustedIndex = dimensionHeadSecond * decreasing + decreasing

This pattern appears when converting between coordinate systems or computing reverse-indexed
positions with additional boundary adjustments.

You can use `decreasing` as the step parameter in `range` [1] to iterate backward:

>>> for leaf in range(leavesTotal + decreasing, 1, decreasing):
...     processLeaf(leaf)

You can combine `decreasing` with other semantic constants:

>>> boundaryRange = range(
...     (start + inclusive) * decreasing,
...     (stop + inclusive) * decreasing,
...     decreasing
... )

References
----------
[1] Built-in Functions - `range` (Python documentation)
	https://docs.python.org/3/library/functions.html#func-range

"""

errorL33T: int = 31212012
"""Signal an error state with a visually distinctive numeric value.

(AI generated docstring)

The identifier `errorL33T` holds the value `31212012` and serves as a semantic replacement
for typical error values such as `-1`. You can use `errorL33T` wherever an error sentinel
value appears but where the meaning "error state" is more important than the specific
numeric value. The value `31212012` spells "ERROR" in leetspeak [1], making `errorL33T`
more visually distinctive during debugging than common error values like `-1`.

You can use `errorL33T` directly as a positive value to signal an invalid or uninitialized
state. You can use `-errorL33T` when a negative value provides a stronger visual signal
or when the domain expects negative error indicators. The leetspeak encoding makes
`errorL33T` immediately recognizable in debugger output, log files, or printed state,
reducing the risk of overlooking error conditions that might be missed with generic
values like `-1`.

In rare cases when code execution does not raise `Exception` [2] but produces incorrect
output, seeing `31212012` or `-31212012` in the output provides a stronger diagnostic
signal than `-1` or other conventional error values. The distinctive value helps identify
which variables hold error states without requiring careful inspection of every numeric
field.

Examples
--------
You can use `errorL33T` as an initial value for variables that must be assigned before use:

>>> countTotal: int = -errorL33T
>>> # ... computation logic that assigns countTotal ...
>>> if countTotal == -errorL33T:
...     raise ValueError("countTotal was never assigned")

You can use `errorL33T` to mark invalid index positions:

>>> pileOf_kCrease: Pile = errorL33T
>>> pileOf_rCrease: Pile = errorL33T
>>> # ... logic that assigns valid pile indices ...
>>> if pileOf_kCrease == errorL33T:
...     raise ValueError("pileOf_kCrease remains uninitialized")

You can use `-errorL33T` when negative values signal forbidden states:

>>> leafForbidden: Leaf = -errorL33T
>>> # ... validation logic ...
>>> if leafForbidden == -errorL33T:
...     applyDefaultBehavior()

References
----------
[1] Leet - Wikipedia
	https://en.wikipedia.org/wiki/Leet
[2] Built-in Exceptions - `Exception` (Python documentation)
	https://docs.python.org/3/library/exceptions.html#Exception

"""

inclusive: int = 1
"""Express inclusion (or exclusion) of a boundary value.

(AI generated docstring)

The identifier `inclusive` holds the value `1` and serves as a semantic replacement for
the numeric literal `1` when adjusting boundary computations. You can use `inclusive` to
convert Python's default half-open interval semantics `[start, stop)` into closed intervals
`[start, stop]` by adding `inclusive` to the upper bound. You can also use `- inclusive`
to signal explicit exclusion of a boundary value.

You can use `inclusive` wherever `1` would appear but where the meaning "include this
boundary" or "extend by one position" is more important than the specific numeric value.
Using `inclusive` makes the code's intent explicit: the adjustment exists to change interval
semantics, not to perform arbitrary arithmetic.

Common contexts include: `range` objects [1], `slice` objects [2], sequence indexing,
and any function that accepts boundary parameters with half-open semantics (such as
`random.randrange` [3], `numpy.arange` [4], or `pandas.RangeIndex` [5]).

Many functions in Python packages use half-open intervals by default. For example,
`random.randrange(start, stop)` [3] excludes `stop`, while `random.randint(a, b)` [3]
includes `b`. You can use `inclusive` to make the adjustment explicit when working with
half-open functions:

- `range(1, lastValue + inclusive)` includes `lastValue` in the iteration.
- `slice(start, stop + inclusive)` includes the element at index `stop`.
- `array[start : stop + inclusive]` includes `array[stop]` in the slice.
- `boundary - inclusive` explicitly excludes `boundary` from consideration.

Examples
--------
You can use `inclusive` to extend `range` [1] objects to include the final value:

>>> for leaf1ndex in range(1, leavesTotal + inclusive):
...     processLeaf(leaf1ndex)

Without `inclusive`, `range(1, leavesTotal)` would stop before processing the leaf at
index `leavesTotal`. Adding `inclusive` ensures the loop processes all leaves from
`1` through `leavesTotal`.

You can use `inclusive` in nested `range` [1] objects:

>>> for activeLeaf1ndex in range(1, leavesTotal + inclusive):
...     for connectee1ndex in range(1, activeLeaf1ndex + inclusive):
...         processConnection(activeLeaf1ndex, connectee1ndex)

You can use `inclusive` when constructing `frozenset` [6] objects from `range` [1]:

>>> pilesOpen: frozenset[int] = frozenset(
...     range(pileLast + inclusive)
... ).difference(leavesPinned.keys())

You can use `inclusive` to construct forbidden index ranges:

>>> pilesForbidden = frozenset([
...     *range(pileOf_k),
...     *range(pileOf_kCrease + 1, pileLast + inclusive)
... ])

You can use `- inclusive` to signal explicit exclusion of a boundary:

>>> for pile in filter(filterPredicate, range(0, pile_k - inclusive)):
...     processPile(pile)

In this example, `pile_k - inclusive` explicitly excludes the pile at index `pile_k` from
the range. Filters or predicates that accept inclusive upper bounds (such as `between`-style
functions) may require subtracting `inclusive` to express exclusive boundaries.

You can combine `inclusive` with other semantic constants:

>>> listIndicesPilesExcluded.extend(
...     range(
...         (1 + inclusive) * decreasing,
...         (stop + inclusive) * decreasing,
...         decreasing
...     )
... )

References
----------
[1] Built-in Functions - `range` (Python documentation)
	https://docs.python.org/3/library/functions.html#func-range
[2] Built-in Types - `slice` (Python documentation)
	https://docs.python.org/3/library/functions.html#slice
[3] `random` - Generate pseudo-random numbers (Python documentation)
	https://docs.python.org/3/library/random.html
[4] `numpy.arange` (NumPy documentation)
	https://numpy.org/doc/stable/reference/generated/numpy.arange.html
[5] `pandas.RangeIndex` (pandas documentation)
	https://pandas.pydata.org/docs/reference/api/pandas.RangeIndex.html
[6] Built-in Types - `frozenset` (Python documentation)
	https://docs.python.org/3/library/stdtypes.html#frozenset

"""

zeroIndexed: int = 1
"""Express that the adjustment to a value is due to zero-based indexing.

(AI generated docstring)

The identifier `zeroIndexed` holds the value `1` and serves as a semantic replacement
for the numeric literal `1` when converting between zero-based indexing (Python's default)
and one-based indexing (common in mathematical notation, human-readable numbering, and
many domain-specific conventions). You can use `zeroIndexed` as an addend or subtrahend
to adjust index values, counts, or boundary computations.

You can use `zeroIndexed` wherever `1` would appear but where the meaning "adjust for
indexing convention" is more important than the specific numeric value. Using `zeroIndexed`
makes the code's intent explicit: the adjustment exists to reconcile indexing systems, not
to perform arbitrary arithmetic.

The most common usage is `count - zeroIndexed` to convert a one-based count ("there are N
items") to the zero-based index of the last item (`N - 1`). You can also use `+ zeroIndexed`
when converting from zero-based indices back to one-based positions or counts, or when
adjusting formulas that assume one-based indexing.

Common contexts include: computing final indices from counts, converting between mathematical
notation (often one-based) and Python code (zero-based), accessing the last valid index of
a sequence, and boundary calculations in algorithms that mix indexing conventions.

Examples
--------
You can use `- zeroIndexed` to convert a count to the last valid zero-based index:

>>> lastDimensionIndex = dimensionsTotal - zeroIndexed

If `dimensionsTotal` is `3` (representing three dimensions), then `lastDimensionIndex`
becomes `2`, which correctly identifies the index of the third dimension in a zero-indexed
array `[0, 1, 2]`.

You can use `- zeroIndexed` when accessing elements by position:

>>> voodooMath: int = creaseAnteAt二Ante首[
...     largestPossibleLengthOfListOfCreases - zeroIndexed
... ]

Here, `largestPossibleLengthOfListOfCreases` represents a count ("how many elements"),
and subtracting `zeroIndexed` produces the index of the last element.

You can use `- zeroIndexed` in `range` [1] boundary computations:

>>> listIndicesCreasePostToKeep.extend(
...     range(
...         dimensionsTotal - dimensionHead + 1,
...         dimensionsTotal - zeroIndexed
...     )
... )

You can use `- zeroIndexed` in conditional expressions:

>>> if dimensionsTotal - zeroIndexed - dimensionHead == zerosAtThe首:
...     applySpecialCase()

You can use `+ zeroIndexed` when the adjustment goes in the opposite direction:

>>> productsOfDimensionsTruncator: int = (
...     dimensionFrom首 - (dimensionsTotal + zeroIndexed)
... )

In this example, `dimensionsTotal + zeroIndexed` adjusts the total count upward before
subtraction, compensating for a formula that expects one-based indexing.

References
----------
[1] Built-in Functions - `range` (Python documentation)
	https://docs.python.org/3/library/functions.html#func-range

"""

#======== Some colors for printing text to the terminal. ========
# Many values and options at https://gist.github.com/JBlond/2fea43a3049b38287e5e9cefc87b2124
# Many, many, many options at https://stackoverflow.com/a/33206814/4403878
# NOTE Always define color and background color at the same time.

ansiColorReset: str = '\x1b[0m'
"""Reset terminal text color and background to default settings.

(AI generated docstring)

The identifier `ansiColorReset` holds the ANSI escape sequence [1] `\x1b[0m` that resets
terminal foreground and background colors to their default values. You can use
`ansiColorReset` after printing colored text to ensure subsequent output appears in the
terminal's default color scheme.

ANSI escape sequences [1] allow programs to control terminal display attributes by
embedding special character sequences in output text. The sequence `\x1b[0m` (ESC[0m)
resets all text attributes, including color, weight, and style, returning the terminal
to its default state.

You can concatenate `ansiColorReset` with output strings or write `ansiColorReset` as
a separate operation after colored text. Failing to reset colors causes subsequent
output to inherit the previous color settings, which can make output difficult to read
if default-colored text appears against a colored background.

Most modern terminal emulators support ANSI escape sequences [1] on all platforms,
including Windows Terminal, macOS Terminal, and Linux terminals. Legacy Windows console
applications may require enabling ANSI support through Windows API calls or environment
variables.

Examples
--------
You can use `ansiColorReset` to reset colors after printing colored output:

>>> import sys
>>> sys.stdout.write(f"{ansiColors.GreenOnBlack}Success{ansiColorReset}\n")

You can use `ansiColorReset` in f-strings for inline color resets:

>>> message = f"{ansiColors.YellowOnRed}Error{ansiColorReset}: Invalid input"
>>> print(message)

You can use `ansiColorReset` after multiple colored segments:

>>> sys.stdout.write(f"{ansiColors.WhiteOnBlue}Header{ansiColorReset}\n")
>>> # ... more output ...
>>> sys.stdout.write(ansiColorReset)  # Ensure reset before exit

References
----------
[1] ANSI escape code - Wikipedia
	https://en.wikipedia.org/wiki/ANSI_escape_code

"""

class AnsiColors(NamedTuple):
	r"""Provide high-contrast ANSI color combinations for terminal output.

	(AI generated docstring)

	You can use this class to access ANSI escape sequences [1] for displaying colored
	text in terminal emulators. Each attribute holds an escape sequence that sets both
	foreground and background colors simultaneously. The color combinations were selected
	through research into high-contrast color visibility [2] to maximize readability
	across different terminal color schemes and viewing conditions.

	ANSI escape sequences [1] control terminal display attributes by embedding special
	character sequences in output text. The sequences in `AnsiColors` use the format
	`\\x1b[foreground;backgroundm` where foreground and background are numeric color codes.
	Each sequence sets both colors simultaneously to ensure sufficient contrast regardless
	of the terminal's default color scheme.

	You can access color combinations through attribute names following the pattern
	`ForegroundOnBackground`, such as `BlackOnCyan` or `WhiteOnRed`. The colors were
	chosen to provide strong visual contrast while remaining comfortable to read during
	extended debugging sessions. Always use `ansiColorReset` [3] after colored output
	to return the terminal to default colors.

	The `NamedTuple` [4] base class makes `AnsiColors` immutable and provides both
	attribute access and sequence indexing. You can access colors by name
	(`ansiColors.GreenOnBlack`) or by index (`ansiColors[0]`), enabling dynamic color
	selection based on computed indices.

	Attributes
	----------
	BlackOnCyan : str
		ANSI escape sequence for black text on cyan background.
	BlackOnMagenta : str
		ANSI escape sequence for black text on magenta background.
	BlackOnWhite : str
		ANSI escape sequence for black text on white background.
	BlackOnYellow : str
		ANSI escape sequence for black text on yellow background.
	BlueOnWhite : str
		ANSI escape sequence for blue text on white background.
	BlueOnYellow : str
		ANSI escape sequence for blue text on yellow background.
	CyanOnBlack : str
		ANSI escape sequence for cyan text on black background.
	CyanOnBlue : str
		ANSI escape sequence for cyan text on blue background.
	CyanOnMagenta : str
		ANSI escape sequence for cyan text on magenta background.
	GreenOnBlack : str
		ANSI escape sequence for green text on black background.
	MagentaOnBlack : str
		ANSI escape sequence for magenta text on black background.
	MagentaOnBlue : str
		ANSI escape sequence for magenta text on blue background.
	MagentaOnCyan : str
		ANSI escape sequence for magenta text on cyan background.
	RedOnWhite : str
		ANSI escape sequence for red text on white background.
	WhiteOnBlack : str
		ANSI escape sequence for white text on black background.
	WhiteOnBlue : str
		ANSI escape sequence for white text on blue background.
	WhiteOnMagenta : str
		ANSI escape sequence for white text on magenta background.
	WhiteOnRed : str
		ANSI escape sequence for white text on red background.
	YellowOnBlack : str
		ANSI escape sequence for yellow text on black background.
	YellowOnBlue : str
		ANSI escape sequence for yellow text on blue background.
	YellowOnRed : str
		ANSI escape sequence for yellow text on red background.

	Examples
	--------
	You can access color combinations by attribute name:

	>>> import sys
	>>> sys.stdout.write(f"{ansiColors.GreenOnBlack}Success{ansiColorReset}\\n")

	You can use boolean indexing to select colors based on conditions:

	>>> matchColor = (ansiColors.YellowOnRed, ansiColors.GreenOnBlack)[resultMatches]
	>>> sys.stdout.write(f"{matchColor}{result}{ansiColorReset}\\n")

	You can use computed indices to select colors dynamically:

	>>> colorIndex = int(identifier, 36) % len(ansiColors)
	>>> sys.stdout.write(f"{ansiColors[colorIndex]}{identifier}{ansiColorReset}")

	This pattern provides consistent color assignment for identifiers based on their
	string representation, making repeated identifiers appear in the same color across
	multiple runs.

	You can combine multiple colored segments in output:

	>>> sys.stdout.write(
	...     f"{ansiColors.WhiteOnBlue}Status: {ansiColorReset}"
	...     f"{ansiColors.GreenOnBlack}Complete{ansiColorReset}\\n"
	... )

	References
	----------
	[1] ANSI escape code - Wikipedia
		https://en.wikipedia.org/wiki/ANSI_escape_code
	[2] Color contrast - Wikipedia
		https://en.wikipedia.org/wiki/Contrast_(vision)#Color_contrast
	[3] ansiColorReset
		Internal package reference
	[4] Built-in Types - `typing.NamedTuple` (Python documentation)
		https://docs.python.org/3/library/typing.html#typing.NamedTuple
	"""

	BlackOnCyan: str = '\x1b[30;46m'
	BlackOnMagenta: str = '\x1b[30;45m'
	BlackOnWhite: str = '\x1b[30;47m'
	BlackOnYellow: str = '\x1b[30;43m'
	BlueOnWhite: str = '\x1b[34;47m'
	BlueOnYellow: str = '\x1b[34;43m'
	CyanOnBlack: str = '\x1b[36;40m'
	CyanOnBlue: str = '\x1b[36;44m'
	CyanOnMagenta: str = '\x1b[36;45m'
	GreenOnBlack: str = '\x1b[32;40m'
	MagentaOnBlack: str = '\x1b[35;40m'
	MagentaOnBlue: str = '\x1b[35;44m'
	MagentaOnCyan: str = '\x1b[35;46m'
	RedOnWhite: str = '\x1b[31;47m'
	WhiteOnBlack: str = '\x1b[37;40m'
	WhiteOnBlue: str = '\x1b[37;44m'
	WhiteOnMagenta: str = '\x1b[37;45m'
	WhiteOnRed: str = '\x1b[37;41m'
	YellowOnBlack: str = '\x1b[33;40m'
	YellowOnBlue: str = '\x1b[33;44m'
	YellowOnRed: str = '\x1b[33;41m'




from .. import dicttoolz as _dicttoolz, itertoolz as _itertoolz, recipes as _recipes
from ..functoolz import (
	apply as apply, complement as complement, compose as comp, compose as compose, compose_left as compose_left, curry as curry,
	excepts as _excepts_class, flip as flip, identity as identity, juxt as juxt, memoize as memoize, pipe as pipe,
	thread_first as thread_first, thread_last as thread_last)
from ..itertoolz import (
	concat as concat, concatv as concatv, count as count, diff as diff, first as first, frequencies as frequencies, interleave as interleave,
	isdistinct as isdistinct, isiterable as isiterable, last as last, merge_sorted as merge_sorted, peek as peek, second as second)
# All functions from operator module are re-exported here.
# Binary and n-ary functions are curried; unary functions are not.
# From a typing perspective, curried functions have identical signatures.
from . import operator
from .exceptions import merge, merge_with
from _typeshed import SupportsRichComparison
from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping, MutableMapping, Sequence
from typing import Any, Literal, overload, TypeGuard
from typing_extensions import TypeIs
import functools

__all__ = [
	# Curried functions (defined in this module)
	"accumulate",
	"assoc",
	"assoc_in",
	"cons",
	"countby",
	"dissoc",
	"do",
	"drop",
	"excepts",
	"filter",
	"get",
	"get_in",
	"groupby",
	"interpose",
	"itemfilter",
	"itemmap",
	"iterate",
	"join",
	"keyfilter",
	"keymap",
	"map",
	"mapcat",
	"nth",
	"partial",
	"partition",
	"partition_all",
	"partitionby",
	"peekn",
	"pluck",
	"random_sample",
	"reduce",
	"reduceby",
	"remove",
	"sliding_window",
	"sorted",
	"tail",
	"take",
	"take_nth",
	"topk",
	"unique",
	"update_in",
	"valfilter",
	"valmap",
	# Re-exported (not curried)
	"apply",
	"comp",
	"complement",
	"compose",
	"compose_left",
	"concat",
	"concatv",
	"count",
	"curry",
	"diff",
	"first",
	"flip",
	"frequencies",
	"identity",
	"interleave",
	"isdistinct",
	"isiterable",
	"juxt",
	"last",
	"memoize",
	"merge_sorted",
	"peek",
	"pipe",
	"second",
	"thread_first",
	"thread_last",
	# Re-exported from .exceptions
	"merge",
	"merge_with",
	# Submodule
	"operator",
]

# Curried accumulate with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def accumulate[T]() -> Callable[..., Iterator[T]]: ...

# Stage 1: Just binop - returns callable waiting for seq (and optional initial)
@overload
def accumulate[T](
	binop: Callable[[T, T], T], /
) -> Callable[..., Iterator[T]]: ...

# Stage 2a: binop + seq (no initial) - executes immediately
@overload
def accumulate[T](
	binop: Callable[[T, T], T], seq: Iterable[T], /
) -> Iterator[T]: ...

# Stage 2b: binop + seq + initial - executes immediately
@overload
def accumulate[T](
	binop: Callable[[T, T], T],
	seq: Iterable[T],
	initial: T,
	/,
) -> Iterator[T]: ...
def accumulate[T](
	binop: Callable[[T, T], T] = ...,
	seq: Iterable[T] = ...,
	initial: T = ...,
) -> Iterator[T] | Callable[..., Iterator[T]]:
	...

@overload
def assoc[K, V]() -> Callable[
	..., dict[K, V] | MutableMapping[K, V]
]: ...
@overload
def assoc[K, V](
	d: Mapping[K, V], /
) -> Callable[..., dict[K, V] | MutableMapping[K, V]]: ...
@overload
def assoc[K, V](
	d: Mapping[K, V], key: K, /
) -> Callable[[V], dict[K, V]]: ...
@overload
def assoc[K, V](
	d: Mapping[K, V],
	key: K,
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> Callable[[V], MutableMapping[K, V]]: ...
@overload
def assoc[K, V](
	d: Mapping[K, V], key: K, value: V, /
) -> dict[K, V]: ...
@overload
def assoc[K, V](
	d: Mapping[K, V],
	key: K,
	value: V,
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> MutableMapping[K, V]: ...
def assoc[K, V](
	d: Mapping[K, V] = ...,
	key: K = ...,
	value: V = ...,
	*,
	factory: Callable[[], MutableMapping[K, V]] = dict,
) -> (
	dict[K, V]
	| MutableMapping[K, V]
	| Callable[..., dict[K, V] | MutableMapping[K, V]]
):
	...

assoc_in = curry(_dicttoolz.assoc_in)

# Curried cons with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def cons[T]() -> Callable[..., Iterator[T]]: ...

# Stage 1: Just el - returns callable waiting for seq
@overload
def cons[T](
	el: T, /
) -> Callable[[Iterable[T]], Iterator[T]]: ...

# Stage 2: Full application - executes immediately
@overload
def cons[T](
	el: T, seq: Iterable[T], /
) -> Iterator[T]: ...
def cons[T](
	el: T = ..., seq: Iterable[T] = ...
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

countby = curry(_recipes.countby)
dissoc = curry(_dicttoolz.dissoc)

# Curried do with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def do[T]() -> Callable[..., T]: ...

# Stage 1: Just func - returns callable waiting for x
@overload
def do[T](func: Callable[[T], Any], /) -> Callable[[T], T]: ...

# Stage 2: Full application - executes immediately
@overload
def do[T](func: Callable[[T], Any], x: T, /) -> T: ...
def do[T](
	func: Callable[[T], Any] = ..., x: T = ...
) -> T | Callable[[T], T] | Callable[..., T]:
	...

@overload
def drop[T]() -> Callable[..., Iterator[T]]: ...
@overload
def drop[T](
	n: int, /
) -> Callable[[Iterable[T]], Iterator[T]]: ...
@overload
def drop[T](
	n: int, seq: Iterable[T], /
) -> Iterator[T]: ...
def drop[T](
	n: int = ..., seq: Iterable[T] = ...
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

@overload
def excepts[T, **P]() -> Callable[..., _excepts_class[T, P]]: ...
@overload
def excepts[T, **P](
	exc: type[Exception] | tuple[type[Exception], ...], /
) -> (
	Callable[[Callable[P, T]], _excepts_class[T, P]]
	| Callable[
		[Callable[P, T], Callable[[Exception], T]],
		_excepts_class[T, P],
	]
): ...
@overload
def excepts[T, **P](
	exc: type[Exception] | tuple[type[Exception], ...],
	func: Callable[P, T],
	/,
) -> _excepts_class[T, P]: ...
@overload
def excepts[T, **P](
	exc: type[Exception] | tuple[type[Exception], ...],
	func: Callable[P, T],
	handler: Callable[[Exception], T],
	/,
) -> _excepts_class[T, P]: ...
def excepts[T, **P](
	exc: type[Exception] | tuple[type[Exception], ...] = ...,
	func: Callable[P, T] = ...,
	handler: Callable[[Exception], T] | None = ...,
) -> _excepts_class[T, P] | Callable[..., _excepts_class[T, P]]:
	...

@overload
def filter[T]() -> Callable[  # noqa: A001
	..., Iterator[T] | Callable[..., Iterator[T]]
]: ...
@overload
def filter[T](  # noqa: A001
	function: None, /
) -> Callable[
	[Iterable[T | None]], Iterator[T]
]: ...
@overload
def filter[S, T](
	function: Callable[[S], TypeGuard[T]], /
) -> Callable[[Iterable[S]], Iterator[T]]: ...
@overload
def filter[S, T](
	function: Callable[[S], TypeIs[T]], /
) -> Callable[[Iterable[S]], Iterator[T]]: ...
@overload
def filter[T](
	function: Callable[[T], Any], /
) -> Callable[[Iterable[T]], Iterator[T]]: ...
@overload
def filter[T](
	function: None, iterable: Iterable[T | None], /
) -> Iterator[T]: ...
@overload
def filter[S, T](
	function: Callable[[S], TypeGuard[T]],
	iterable: Iterable[S],
	/,
) -> Iterator[T]: ...
@overload
def filter[S, T](
	function: Callable[[S], TypeIs[T]],
	iterable: Iterable[S],
	/,
) -> Iterator[T]: ...
@overload
def filter[T](
	function: Callable[[T], Any],
	iterable: Iterable[T],
	/,
) -> Iterator[T]: ...
def filter[T](
	function: Callable[[T], Any] | None = ...,
	iterable: Iterable[T] = ...,
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[
		...,
		Iterator[T] | Callable[..., Iterator[T]],
	]
):
	...

@overload
def get[T]() -> Callable[..., T | tuple[T, ...]]: ...
@overload
def get[T](
	ind: Sequence[Any], /
) -> (
	Callable[
		[Sequence[T] | Mapping[Any, T]],
		tuple[T, ...],
	]
	| Callable[
		[Sequence[T] | Mapping[Any, T], T],
		tuple[T, ...],
	]
): ...
@overload
def get[T](
	ind: Any, /
) -> (
	Callable[
		[Sequence[T] | Mapping[Any, T]], T
	]
	| Callable[
		[Sequence[T] | Mapping[Any, T], T], T
	]
): ...
@overload
def get[T](
	ind: Sequence[Any],
	seq: Sequence[T] | Mapping[Any, T],
	/,
) -> tuple[T, ...]: ...
@overload
def get[T](
	ind: Any,
	seq: Sequence[T] | Mapping[Any, T],
	/,
) -> T: ...
@overload
def get[T](
	ind: Sequence[Any],
	seq: Sequence[T] | Mapping[Any, T],
	default: T,
	/,
) -> tuple[T, ...]: ...
@overload
def get[T](
	ind: Any,
	seq: Sequence[T] | Mapping[Any, T],
	default: T,
	/,
) -> T: ...
def get[T](
	ind: Any | Sequence[Any] = ...,
	seq: Sequence[T] | Mapping[Any, T] = ...,
	default: T = ...,
) -> T | tuple[T, ...] | Callable[..., T | tuple[T, ...]]:
	...

get_in = curry(_dicttoolz.get_in)

@overload
def groupby[KT, T]() -> Callable[..., dict[KT, list[T]]]: ...
@overload
def groupby[KT, T](
	key: Callable[[T], KT], /
) -> Callable[[Iterable[T]], dict[KT, list[T]]]: ...
@overload
def groupby[T](
	key: Any, /
) -> Callable[[Iterable[T]], dict[Any, list[T]]]: ...
@overload
def groupby[KT, T](
	key: Callable[[T], KT], seq: Iterable[T], /
) -> dict[KT, list[T]]: ...
@overload
def groupby[T](
	key: Any, seq: Iterable[T], /
) -> dict[Any, list[T]]: ...
def groupby[KT, T](
	key: Callable[[T], KT] | Any = ...,
	seq: Iterable[T] = ...,
) -> (
	dict[KT, list[T]]
	| dict[Any, list[T]]
	| Callable[..., dict[KT, list[T]] | dict[Any, list[T]]]
):
	...

# Curried interpose with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def interpose[T]() -> Callable[..., Iterator[T]]: ...

# Stage 1: Just el - returns callable waiting for seq
@overload
def interpose[T](
	el: T, /
) -> Callable[[Iterable[T]], Iterator[T]]: ...

# Stage 2: Full application - executes immediately
@overload
def interpose[T](
	el: T, seq: Iterable[T], /
) -> Iterator[T]: ...
def interpose[T](
	el: T = ..., seq: Iterable[T] = ...
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

@overload
def itemfilter[K, V]() -> Callable[
	..., dict[K, V] | MutableMapping[K, V]
]: ...

# Stage 1a: Just predicate (no factory) - returns callable waiting for dict
@overload
def itemfilter[K, V](
	predicate: Callable[[tuple[K, V]], bool], /
) -> Callable[[Mapping[K, V]], dict[K, V]]: ...

# Stage 1b: Predicate with factory - returns callable waiting for dict
@overload
def itemfilter[K, V](
	predicate: Callable[[tuple[K, V]], bool],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> Callable[
	[Mapping[K, V]], MutableMapping[K, V]
]: ...

# Stage 2a: Full application (no factory) - executes immediately
@overload
def itemfilter[K, V](
	predicate: Callable[[tuple[K, V]], bool],
	d: Mapping[K, V],
	/,
) -> dict[K, V]: ...

# Stage 2b: Full application (with factory) - executes immediately
@overload
def itemfilter[K, V](
	predicate: Callable[[tuple[K, V]], bool],
	d: Mapping[K, V],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> MutableMapping[K, V]: ...
def itemfilter[K, V](
	predicate: Callable[[tuple[K, V]], bool] = ...,
	d: Mapping[K, V] = ...,
	*,
	factory: Callable[[], MutableMapping[K, V]] = dict,
) -> (
	dict[K, V]
	| MutableMapping[K, V]
	| Callable[..., dict[K, V] | MutableMapping[K, V]]
):
	...

@overload
def itemmap[K0, V0, K1, V1]() -> Callable[
	..., dict[K1, V1] | MutableMapping[K1, V1]
]: ...

# Stage 1a: Just func (no factory) - returns callable waiting for dict
@overload
def itemmap[K0, V0, K1, V1](
	func: Callable[[tuple[K0, V0]], tuple[K1, V1]], /
) -> Callable[[Mapping[K0, V0]], dict[K1, V1]]: ...

# Stage 1b: Func with factory - returns callable waiting for dict
@overload
def itemmap[K0, V0, K1, V1](
	func: Callable[[tuple[K0, V0]], tuple[K1, V1]],
	/,
	*,
	factory: Callable[[], MutableMapping[K1, V1]],
) -> Callable[
	[Mapping[K0, V0]], MutableMapping[K1, V1]
]: ...

# Stage 2a: Full application (no factory) - executes immediately
@overload
def itemmap[K0, V0, K1, V1](
	func: Callable[[tuple[K0, V0]], tuple[K1, V1]],
	d: Mapping[K0, V0],
	/,
) -> dict[K1, V1]: ...

# Stage 2b: Full application (with factory) - executes immediately
@overload
def itemmap[K0, V0, K1, V1](
	func: Callable[[tuple[K0, V0]], tuple[K1, V1]],
	d: Mapping[K0, V0],
	/,
	*,
	factory: Callable[[], MutableMapping[K1, V1]],
) -> MutableMapping[K1, V1]: ...
def itemmap[K0, V0, K1, V1](
	func: Callable[[tuple[K0, V0]], tuple[K1, V1]] = ...,
	d: Mapping[K0, V0] = ...,
	*,
	factory: Callable[[], MutableMapping[K1, V1]] = dict,
) -> (
	dict[K1, V1]
	| MutableMapping[K1, V1]
	| Callable[..., dict[K1, V1] | MutableMapping[K1, V1]]
):
	...

# Curried iterate with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def iterate[T]() -> Callable[..., Iterator[T]]: ...

# Stage 1: Just func - returns callable waiting for x
@overload
def iterate[T](
	func: Callable[[T], T], /
) -> Callable[[T], Iterator[T]]: ...

# Stage 2: Full application - executes immediately
@overload
def iterate[T](
	func: Callable[[T], T], x: T, /
) -> Iterator[T]: ...
def iterate[T](
	func: Callable[[T], T] = ..., x: T = ...
) -> (
	Iterator[T]
	| Callable[[T], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

# Curried join with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def join[T, U]() -> Callable[..., Iterator[tuple[T, U]]]: ...

# Stage 1: Just leftkey - returns a callable
@overload
def join[T, U](
	leftkey: Callable[[T], Hashable], /
) -> Callable[..., Iterator[tuple[T, U]]]: ...

# Stage 2: leftkey + leftseq - returns a callable
@overload
def join[T, U](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	/,
) -> Callable[..., Iterator[tuple[T, U]]]: ...

# Stage 3: leftkey + leftseq + rightkey - returns callable waiting for rightseq
# This is the key overload for pipe usage!
# Note: We use Any for U because U can't be inferred until rightseq is provided.
# The callable will properly infer types when called with rightseq.
@overload
def join[T](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[..., Hashable],
	/,
) -> Callable[
	[Iterable[Any]],
	Iterator[tuple[T, Any]],
]: ...

# Stage 4a: Full application (inner join) - executes immediately
@overload
def join[T, U](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	/,
) -> Iterator[tuple[T, U]]: ...

# Stage 4b: Full application with left_default only (right outer join)
@overload
def join[T, U, L](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	/,
	left_default: L,
) -> Iterator[tuple[T | L, U]]: ...

# Stage 4c: Full application with right_default only (left outer join)
@overload
def join[T, U, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	/,
	*,
	right_default: R,
) -> Iterator[tuple[T, U | R]]: ...

# Stage 4d: Full application with both defaults (full outer join)
@overload
def join[T, U, L, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	/,
	left_default: L,
	right_default: R,
) -> Iterator[tuple[T | L, U | R]]: ...

# Stage 3 with defaults: leftkey + leftseq + rightkey + defaults - returns callable
@overload
def join[T, U, L, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	/,
	left_default: L,
	right_default: R,
) -> Callable[
	[Iterable[U]], Iterator[tuple[T | L, U | R]]
]: ...

# Implementation signature
def join[T, U, L, R](
	leftkey: Callable[[T], Hashable] | Hashable = ...,
	leftseq: Iterable[T] = ...,
	rightkey: Callable[[U], Hashable] | Hashable = ...,
	rightseq: Iterable[U] = ...,
	left_default: L = ...,
	right_default: R = ...,
) -> (
	Iterator[tuple[T | L, U | R]]
	| Callable[..., Iterator[tuple[T | L, U | R]]]
):
	...

# Curried keyfilter with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def keyfilter[K, V]() -> Callable[
	..., dict[K, V] | MutableMapping[K, V]
]: ...

# Stage 1a: Just predicate (no factory) - returns callable waiting for dict
@overload
def keyfilter[K, V](
	predicate: Callable[[K], bool], /
) -> Callable[[Mapping[K, V]], dict[K, V]]: ...

# Stage 1b: Predicate with factory - returns callable waiting for dict
@overload
def keyfilter[K, V](
	predicate: Callable[[K], bool],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> Callable[
	[Mapping[K, V]], MutableMapping[K, V]
]: ...

# Stage 2a: Full application (no factory) - executes immediately
@overload
def keyfilter[K, V](
	predicate: Callable[[K], bool],
	d: Mapping[K, V],
	/,
) -> dict[K, V]: ...

# Stage 2b: Full application (with factory) - executes immediately
@overload
def keyfilter[K, V](
	predicate: Callable[[K], bool],
	d: Mapping[K, V],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> MutableMapping[K, V]: ...
def keyfilter[K, V](
	predicate: Callable[[K], bool] = ...,
	d: Mapping[K, V] = ...,
	*,
	factory: Callable[[], MutableMapping[K, V]] = dict,
) -> (
	dict[K, V]
	| MutableMapping[K, V]
	| Callable[..., dict[K, V] | MutableMapping[K, V]]
):
	...

@overload
def keymap[K0, K1, V]() -> Callable[
	..., dict[K1, V] | MutableMapping[K1, V]
]: ...

# Stage 1a: Just func (no factory) - returns callable waiting for dict
@overload
def keymap[K0, K1, V](
	func: Callable[[K0], K1], /
) -> Callable[[Mapping[K0, V]], dict[K1, V]]: ...

# Stage 1b: Func with factory - returns callable waiting for dict
@overload
def keymap[K0, K1, V](
	func: Callable[[K0], K1],
	/,
	*,
	factory: Callable[[], MutableMapping[K1, V]],
) -> Callable[
	[Mapping[K0, V]], MutableMapping[K1, V]
]: ...

# Stage 2a: Full application (no factory) - executes immediately
@overload
def keymap[K0, K1, V](
	func: Callable[[K0], K1],
	d: Mapping[K0, V],
	/,
) -> dict[K1, V]: ...

# Stage 2b: Full application (with factory) - executes immediately
@overload
def keymap[K0, K1, V](
	func: Callable[[K0], K1],
	d: Mapping[K0, V],
	/,
	*,
	factory: Callable[[], MutableMapping[K1, V]],
) -> MutableMapping[K1, V]: ...
def keymap[K0, K1, V](
	func: Callable[[K0], K1] = ...,
	d: Mapping[K0, V] = ...,
	*,
	factory: Callable[[], MutableMapping[K1, V]] = dict,
) -> (
	dict[K1, V]
	| MutableMapping[K1, V]
	| Callable[..., dict[K1, V] | MutableMapping[K1, V]]
):
	...

@overload
def map[T1, S]() -> Callable[
	..., Iterator[S] | Callable[..., Iterator[S]]
]: ...
@overload
def map[T1, S](
	func: Callable[[T1], S], /
) -> Callable[[Iterable[T1]], Iterator[S]]: ...
@overload
def map[T1, T2, S](
	func: Callable[[T1, T2], S], /
) -> Callable[
	[Iterable[T1], Iterable[T2]],
	Iterator[S],
]: ...
@overload
def map[T1, T2, T3, S](
	func: Callable[[T1, T2, T3], S], /
) -> Callable[
	[
		Iterable[T1],
		Iterable[T2],
		Iterable[T3],
	],
	Iterator[S],
]: ...
@overload
def map[T1, T2, T3, T4, S](
	func: Callable[[T1, T2, T3, T4], S], /
) -> Callable[
	[
		Iterable[T1],
		Iterable[T2],
		Iterable[T3],
		Iterable[T4],
	],
	Iterator[S],
]: ...
@overload
def map[T1, T2, T3, T4, T5, S](
	func: Callable[[T1, T2, T3, T4, T5], S], /
) -> Callable[
	[
		Iterable[T1],
		Iterable[T2],
		Iterable[T3],
		Iterable[T4],
		Iterable[T5],
	],
	Iterator[S],
]: ...
@overload
def map[T1, S](
	func: Callable[[T1], S], iterable: Iterable[T1], /
) -> Iterator[S]: ...
@overload
def map[T1, T2, S](
	func: Callable[[T1, T2], S],
	iterable: Iterable[T1],
	iter2: Iterable[T2],
	/,
) -> Iterator[S]: ...
@overload
def map[T1, T2, T3, S](
	func: Callable[[T1, T2, T3], S],
	iterable: Iterable[T1],
	iter2: Iterable[T2],
	iter3: Iterable[T3],
	/,
) -> Iterator[S]: ...
@overload
def map[T1, T2, T3, T4, S](
	func: Callable[[T1, T2, T3, T4], S],
	iterable: Iterable[T1],
	iter2: Iterable[T2],
	iter3: Iterable[T3],
	iter4: Iterable[T4],
	/,
) -> Iterator[S]: ...
@overload
def map[T1, T2, T3, T4, T5, S](
	func: Callable[[T1, T2, T3, T4, T5], S],
	iterable: Iterable[T1],
	iter2: Iterable[T2],
	iter3: Iterable[T3],
	iter4: Iterable[T4],
	iter5: Iterable[T5],
	/,
) -> Iterator[S]: ...
@overload
def map[S](
	func: Callable[..., S],
	iterable: Iterable[Any],
	iter2: Iterable[Any],
	iter3: Iterable[Any],
	iter4: Iterable[Any],
	iter5: Iterable[Any],
	iter6: Iterable[Any],
	/,
	*iterables: Iterable[Any],
) -> Iterator[S]: ...
def map[S](
	func: Callable[..., S] = ...,
	*iterables: Iterable[Any],
) -> (
	Iterator[S]
	| Callable[..., Iterator[S]]
	| Callable[
		...,
		Iterator[S] | Callable[..., Iterator[S]],
	]
):
	...

@overload
def mapcat[T, R]() -> Callable[
	..., Iterator[R] | Callable[..., Iterator[R]]
]: ...
@overload
def mapcat[T, R](
	func: Callable[[T], Iterable[R]], /
) -> Callable[[Iterable[T]], Iterator[R]]: ...
@overload
def mapcat[T, R](
	func: Callable[[T], Iterable[R]],
	seqs: Iterable[T],
	/,
) -> Iterator[R]: ...
def mapcat[T, R](
	func: Callable[[T], Iterable[R]] = ...,
	seqs: Iterable[T] = ...,
) -> (
	Iterator[R]
	| Callable[[Iterable[T]], Iterator[R]]
	| Callable[
		...,
		Iterator[R] | Callable[..., Iterator[R]],
	]
):
	...

# Curried nth with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def nth[T]() -> Callable[..., T]: ...

# Stage 1: Just n - returns callable waiting for seq
@overload
def nth[T](n: int, /) -> Callable[[Iterable[T]], T]: ...

# Stage 2: Full application - executes immediately
@overload
def nth[T](n: int, seq: Iterable[T], /) -> T: ...
def nth[T](
	n: int = ..., seq: Iterable[T] = ...
) -> T | Callable[[Iterable[T]], T] | Callable[..., T]:
	...

partial = curry(functools.partial)

@overload
def partition[T]() -> Callable[..., Iterator[tuple[T, ...]]]: ...
@overload
def partition[T](
	n: int, /
) -> Callable[..., Iterator[tuple[T, ...]]]: ...
@overload
def partition[T](
	n: Literal[1], seq: Iterable[T], /
) -> Iterator[tuple[T]]: ...
@overload
def partition[T](
	n: int, seq: Iterable[T], /
) -> Iterator[tuple[T, ...]]: ...
@overload
def partition[T](
	n: Literal[1], seq: Iterable[T], pad: Any, /
) -> Iterator[tuple[T]]:
	# Note: With n=1, tuples always have exactly 1 element, so pad is never used
	...

@overload
def partition[T, P](
	n: int, seq: Iterable[T], pad: P, /
) -> Iterator[tuple[T | P, ...]]: ...
def partition[T, P](
	n: int = ...,
	seq: Iterable[T] = ...,
	pad: P = ...,
) -> (
	Iterator[tuple[T, ...]]
	| Iterator[tuple[T | P, ...]]
	| Callable[..., Iterator[tuple[T | P, ...]]]
):
	...

# Curried partition_all with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def partition_all[T]() -> Callable[
	..., Iterator[tuple[T, ...]]
]: ...

# Stage 1: Just n - returns callable waiting for seq
@overload
def partition_all[T](
	n: Literal[1], /
) -> Callable[
	[Iterable[T]], Iterator[tuple[T]]
]: ...
@overload
def partition_all[T](
	n: int, /
) -> Callable[
	[Iterable[T]], Iterator[tuple[T, ...]]
]: ...

# Stage 2: Full application - executes immediately
@overload
def partition_all[T](
	n: Literal[1], seq: Iterable[T], /
) -> Iterator[tuple[T]]: ...
@overload
def partition_all[T](
	n: int, seq: Iterable[T], /
) -> Iterator[tuple[T, ...]]: ...
def partition_all[T](
	n: int = ..., seq: Iterable[T] = ...
) -> (
	Iterator[tuple[T, ...]]
	| Callable[
		[Iterable[T]],
		Iterator[tuple[T, ...]] | Iterator[tuple[T]],
	]
	| Callable[..., Iterator[tuple[T, ...]]]
):
	...

partitionby = curry(_recipes.partitionby)
peekn = curry(_itertoolz.peekn)

@overload
def pluck[T]() -> Callable[
	..., Iterator[T] | Iterator[tuple[T, ...]]
]: ...
@overload
def pluck[T](
	ind: Sequence[Any], /
) -> (
	Callable[
		[
			Iterable[
				Sequence[T] | Mapping[Any, T]
			]
		],
		Iterator[tuple[T, ...]],
	]
	| Callable[
		[
			Iterable[
				Sequence[T] | Mapping[Any, T]
			],
			T,
		],
		Iterator[tuple[T, ...]],
	]
): ...
@overload
def pluck[T](
	ind: Any, /
) -> (
	Callable[
		[
			Iterable[
				Sequence[T] | Mapping[Any, T]
			]
		],
		Iterator[T],
	]
	| Callable[
		[
			Iterable[
				Sequence[T] | Mapping[Any, T]
			],
			T,
		],
		Iterator[T],
	]
): ...
@overload
def pluck[T](
	ind: Sequence[Any],
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	/,
) -> Iterator[tuple[T, ...]]: ...
@overload
def pluck[T](
	ind: Any,
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	/,
) -> Iterator[T]: ...
@overload
def pluck[T](
	ind: Sequence[Any],
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	default: T,
	/,
) -> Iterator[tuple[T, ...]]: ...
@overload
def pluck[T](
	ind: Any,
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	default: T,
	/,
) -> Iterator[T]: ...
def pluck[T](
	ind: Any | Sequence[Any] = ...,
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	] = ...,
	default: T = ...,
) -> (
	Iterator[T]
	| Iterator[tuple[T, ...]]
	| Callable[
		..., Iterator[T] | Iterator[tuple[T, ...]]
	]
):
	...

random_sample = curry(_itertoolz.random_sample)

@overload
def reduce[T]() -> Callable[..., T]: ...
@overload
def reduce[T](
	function: Callable[[T, T], T], /
) -> Callable[[Iterable[T]], T]: ...
@overload
def reduce[T, S](
	function: Callable[[T, S], T], /
) -> Callable[..., T]: ...
@overload
def reduce[T](
	function: Callable[[T, T], T],
	iterable: Iterable[T],
	/,
) -> T: ...
@overload
def reduce[T, S](
	function: Callable[[T, S], T],
	iterable: Iterable[S],
	initial: T,
	/,
) -> T: ...
def reduce[T, S](
	function: Callable[[T, S], T] = ...,
	iterable: Iterable[S] = ...,
	initial: T = ...,
) -> T | Callable[..., T]:
	...

reduceby = curry(_itertoolz.reduceby)

# Curried remove with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def remove[T]() -> Callable[..., Iterable[T]]: ...

# Stage 1: Just predicate - returns callable waiting for seq
@overload
def remove[T](
	predicate: Callable[[T], bool], /
) -> Callable[[Iterable[T]], Iterable[T]]: ...

# Stage 2: Full application - executes immediately
@overload
def remove[T](
	predicate: Callable[[T], bool], seq: Iterable[T], /
) -> Iterable[T]: ...
def remove[T](
	predicate: Callable[[T], bool] = ..., seq: Iterable[T] = ...
) -> (
	Iterable[T]
	| Callable[[Iterable[T]], Iterable[T]]
	| Callable[..., Iterable[T]]
):
	...

# Curried sliding_window with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def sliding_window[T]() -> Callable[
	..., Iterator[tuple[T, ...]]
]: ...

# Stage 1a: Just n=1 - returns callable waiting for seq
@overload
def sliding_window[T](
	n: Literal[1], /
) -> Callable[
	[Iterable[T]], Iterator[tuple[T]]
]: ...

# Stage 1b: Just n=2 - returns callable waiting for seq
@overload
def sliding_window[T](
	n: Literal[2], /
) -> Callable[
	[Iterable[T]], Iterator[tuple[T, T]]
]: ...

# Stage 1c: Just n=3 - returns callable waiting for seq
@overload
def sliding_window[T](
	n: Literal[3], /
) -> Callable[
	[Iterable[T]], Iterator[tuple[T, T, T]]
]: ...

# Stage 1d: Just n (general) - returns callable waiting for seq
@overload
def sliding_window[T](
	n: int, /
) -> Callable[
	[Iterable[T]], Iterator[tuple[T, ...]]
]: ...

# Stage 2a: Full application with n=1 - executes immediately
@overload
def sliding_window[T](
	n: Literal[1], seq: Iterable[T], /
) -> Iterator[tuple[T]]: ...

# Stage 2b: Full application with n=2 - executes immediately
@overload
def sliding_window[T](
	n: Literal[2], seq: Iterable[T], /
) -> Iterator[tuple[T, T]]: ...

# Stage 2c: Full application with n=3 - executes immediately
@overload
def sliding_window[T](
	n: Literal[3], seq: Iterable[T], /
) -> Iterator[tuple[T, T, T]]: ...

# Stage 2d: Full application (general) - executes immediately
@overload
def sliding_window[T](
	n: int, seq: Iterable[T], /
) -> Iterator[tuple[T, ...]]: ...
def sliding_window[T](
	n: int = ..., seq: Iterable[T] = ...
) -> (
	Iterator[tuple[T, ...]]
	| Callable[
		[Iterable[T]], Iterator[tuple[T, ...]]
	]
	| Callable[..., Iterator[tuple[T, ...]]]
):
	...

# Curried sorted with explicit overloads for type safety
# Note: key and reverse are keyword-only parameters in builtin sorted
# Stage 0: No arguments - returns a callable
@overload
def sorted[T]() -> Callable[..., list[T]]: ...

# Stage 1a: Partial application with keyword args only (no key) - returns callable
@overload
def sorted[T](
	*,
	key: None = None,
	reverse: bool = False,
) -> Callable[[Iterable[T]], list[T]]: ...

# Stage 1b: Partial application with keyword args only (with key) - returns callable
@overload
def sorted[T](
	*,
	key: Callable[[T], SupportsRichComparison],
	reverse: bool = False,
) -> Callable[[Iterable[T]], list[T]]: ...

# Stage 2a: Full application (no key) - executes immediately
@overload
def sorted[T](
	iterable: Iterable[T],
	/,
	*,
	key: None = None,
	reverse: bool = False,
) -> list[T]: ...

# Stage 2b: Full application (with key function) - executes immediately
@overload
def sorted[T](
	iterable: Iterable[T],
	/,
	*,
	key: Callable[[T], SupportsRichComparison],
	reverse: bool = False,
) -> list[T]: ...

# Implementation signature (catch-all)
def sorted[T](
	iterable: Iterable[T] = ...,
	/,
	*,
	key: Callable[[T], SupportsRichComparison] | None = None,
	reverse: bool = False,
) -> list[T] | Callable[..., list[T]]:
	...

# Curried tail with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def tail[T]() -> Callable[..., Iterator[T]]: ...

# Stage 1: Just n - returns callable waiting for seq
@overload
def tail[T](
	n: int, /
) -> Callable[[Iterable[T]], Iterator[T]]: ...

# Stage 2: Full application - executes immediately
@overload
def tail[T](
	n: int, seq: Iterable[T], /
) -> Iterator[T]: ...
def tail[T](
	n: int = ..., seq: Iterable[T] = ...
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

@overload
def take[T]() -> Callable[..., Iterator[T]]: ...
@overload
def take[T](
	n: int, /
) -> Callable[[Iterable[T]], Iterator[T]]: ...
@overload
def take[T](
	n: int, seq: Iterable[T], /
) -> Iterator[T]: ...
def take[T](
	n: int = ..., seq: Iterable[T] = ...
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

# Curried take_nth with explicit overloads for type safety
# Stage 0: No arguments - returns a callable
@overload
def take_nth[T]() -> Callable[..., Iterator[T]]: ...

# Stage 1: Just n - returns callable waiting for seq
@overload
def take_nth[T](
	n: int, /
) -> Callable[[Iterable[T]], Iterator[T]]: ...

# Stage 2: Full application - executes immediately
@overload
def take_nth[T](
	n: int, seq: Iterable[T], /
) -> Iterator[T]: ...
def take_nth[T](
	n: int = ..., seq: Iterable[T] = ...
) -> (
	Iterator[T]
	| Callable[[Iterable[T]], Iterator[T]]
	| Callable[..., Iterator[T]]
):
	...

topk = curry(_itertoolz.topk)
unique = curry(_itertoolz.unique)
update_in = curry(_dicttoolz.update_in)

@overload
def valfilter[K, V]() -> Callable[
	..., dict[K, V] | MutableMapping[K, V]
]: ...

# Stage 1a: Just predicate (no factory) - returns callable waiting for dict
@overload
def valfilter[K, V](
	predicate: Callable[[V], bool], /
) -> Callable[[Mapping[K, V]], dict[K, V]]: ...

# Stage 1b: Predicate with factory - returns callable waiting for dict
@overload
def valfilter[K, V](
	predicate: Callable[[V], bool],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> Callable[
	[Mapping[K, V]], MutableMapping[K, V]
]: ...

# Stage 2a: Full application (no factory) - executes immediately
@overload
def valfilter[K, V](
	predicate: Callable[[V], bool],
	d: Mapping[K, V],
	/,
) -> dict[K, V]: ...

# Stage 2b: Full application (with factory) - executes immediately
@overload
def valfilter[K, V](
	predicate: Callable[[V], bool],
	d: Mapping[K, V],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> MutableMapping[K, V]: ...
def valfilter[K, V](
	predicate: Callable[[V], bool] = ...,
	d: Mapping[K, V] = ...,
	*,
	factory: Callable[[], MutableMapping[K, V]] = dict,
) -> (
	dict[K, V]
	| MutableMapping[K, V]
	| Callable[..., dict[K, V] | MutableMapping[K, V]]
):
	...

@overload
def valmap[K, V0, V1]() -> Callable[
	..., dict[K, V1] | MutableMapping[K, V1]
]: ...

# Stage 1a: Just func (no factory) - returns callable waiting for dict
@overload
def valmap[K, V0, V1](
	func: Callable[[V0], V1], /
) -> Callable[[Mapping[K, V0]], dict[K, V1]]: ...

# Stage 1b: Func with factory - returns callable waiting for dict
@overload
def valmap[K, V0, V1](
	func: Callable[[V0], V1],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V1]],
) -> Callable[
	[Mapping[K, V0]], MutableMapping[K, V1]
]: ...

# Stage 2a: Full application (no factory) - executes immediately
@overload
def valmap[K, V0, V1](
	func: Callable[[V0], V1],
	d: Mapping[K, V0],
	/,
) -> dict[K, V1]: ...

# Stage 2b: Full application (with factory) - executes immediately
@overload
def valmap[K, V0, V1](
	func: Callable[[V0], V1],
	d: Mapping[K, V0],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V1]],
) -> MutableMapping[K, V1]: ...
def valmap[K, V0, V1](
	func: Callable[[V0], V1] = ...,
	d: Mapping[K, V0] = ...,
	*,
	factory: Callable[[], MutableMapping[K, V1]] = dict,
) -> (
	dict[K, V1]
	| MutableMapping[K, V1]
	| Callable[..., dict[K, V1] | MutableMapping[K, V1]]
):
	...



from humpy_toolz import countby, first, identity, partitionby

def iseven(x: int) -> bool:
    return x % 2 == 0

def test_countby() -> None:
    assert countby(iseven, [1, 2, 3]) == {True: 1, False: 2}
    assert countby(len, ['cat', 'dog', 'mouse']) == {3: 2, 5: 1}
    assert countby(0, ('ab', 'ac', 'bc')) == {'a': 2, 'b': 1}

def test_partitionby() -> None:
    assert list(partitionby(identity, [])) == []
    vowels = 'aeiou'
    assert list(partitionby(vowels.__contains__, 'abcdefghi')) == [('a',), ('b', 'c', 'd'), ('e',), ('f', 'g', 'h'), ('i',)]
    assert list(map(first, partitionby(identity, [1, 1, 1, 2, 3, 3, 2, 2, 3]))) == [1, 2, 3, 2, 3]
    assert ''.join(map(first, partitionby(identity, 'Khhhaaaaannnnn!!!!'))) == 'Khan!'



from collections.abc import Callable
from types import ModuleType
from humpy_toolz._signatures import builtins
from humpy_toolz.functoolz import (
	curry, has_keywords, has_varargs, is_arity, is_partial_args, is_valid_args, num_required_args)
from humpy_toolz.utils import raises
import functools
import humpy_toolz
import humpy_toolz._signatures as _sigs
import inspect
import itertools
import operator
import sys

def make_func(param_string: str, raise_if_called: bool = True):
    if not param_string.startswith('('):
        param_string = '(%s)' % param_string
    if raise_if_called:
        body = 'raise ValueError("function should not be called")'
    else:
        body = 'return True'
    d: dict[str, object] = {}
    exec(f'def func{param_string}:\n    {body}', globals(), d)
    return d['func']

def test_make_func() -> None:
    f = make_func('')
    assert raises(ValueError, lambda: f())
    assert raises(TypeError, lambda: f(1))
    f = make_func('', raise_if_called=False)
    assert f()
    assert raises(TypeError, lambda: f(1))
    f = make_func('x, y=1', raise_if_called=False)
    assert f(1)
    assert f(x=1)
    assert f(1, 2)
    assert f(x=1, y=2)
    assert raises(TypeError, lambda: f(1, 2, 3))
    f = make_func('(x, y=1)', raise_if_called=False)
    assert f(1)
    assert f(x=1)
    assert f(1, 2)
    assert f(x=1, y=2)
    assert raises(TypeError, lambda: f(1, 2, 3))

def test_is_valid(check_valid=is_valid_args, incomplete: bool = False) -> None:
    orig_check_valid = check_valid
    check_valid = lambda func, *args, **kwargs: orig_check_valid(func, args, kwargs)
    f = make_func('')
    assert check_valid(f)
    assert check_valid(f, 1) is False
    assert check_valid(f, x=1) is False
    f = make_func('x')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1)
    assert check_valid(f, x=1)
    assert check_valid(f, 1, x=2) is False
    assert check_valid(f, 1, y=2) is False
    assert check_valid(f, 1, 2) is False
    assert check_valid(f, x=1, y=2) is False
    f = make_func('x=1')
    assert check_valid(f)
    assert check_valid(f, 1)
    assert check_valid(f, x=1)
    assert check_valid(f, 1, x=2) is False
    assert check_valid(f, 1, y=2) is False
    assert check_valid(f, 1, 2) is False
    assert check_valid(f, x=1, y=2) is False
    f = make_func('*args')
    assert check_valid(f)
    assert check_valid(f, 1)
    assert check_valid(f, 1, 2)
    assert check_valid(f, x=1) is False
    f = make_func('**kwargs')
    assert check_valid(f)
    assert check_valid(f, x=1)
    assert check_valid(f, x=1, y=2)
    assert check_valid(f, 1) is False
    f = make_func('x, *args')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1)
    assert check_valid(f, 1, 2)
    assert check_valid(f, x=1)
    assert check_valid(f, 1, x=1) is False
    assert check_valid(f, 1, y=1) is False
    f = make_func('x, y=1, **kwargs')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1)
    assert check_valid(f, x=1)
    assert check_valid(f, 1, 2)
    assert check_valid(f, x=1, y=2, z=3)
    assert check_valid(f, 1, 2, y=3) is False
    f = make_func('a, b, c=3, d=4')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1) is incomplete
    assert check_valid(f, 1, 2)
    assert check_valid(f, 1, c=3) is incomplete
    assert check_valid(f, 1, e=3) is False
    assert check_valid(f, 1, 2, e=3) is False
    assert check_valid(f, 1, 2, b=3) is False
    assert check_valid(1) is False

def test_is_valid_py3(check_valid=is_valid_args, incomplete: bool = False) -> None:
    orig_check_valid = check_valid
    check_valid = lambda func, *args, **kwargs: orig_check_valid(func, args, kwargs)
    f = make_func('x, *, y=1')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1)
    assert check_valid(f, x=1)
    assert check_valid(f, 1, y=2)
    assert check_valid(f, 1, 2) is False
    assert check_valid(f, 1, z=2) is False
    f = make_func('x, *args, y=1')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1)
    assert check_valid(f, x=1)
    assert check_valid(f, 1, y=2)
    assert check_valid(f, 1, 2, y=2)
    assert check_valid(f, 1, 2)
    assert check_valid(f, 1, z=2) is False
    f = make_func('*, y=1')
    assert check_valid(f)
    assert check_valid(f, 1) is False
    assert check_valid(f, y=1)
    assert check_valid(f, z=1) is False
    f = make_func('x, *, y')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1) is incomplete
    assert check_valid(f, x=1) is incomplete
    assert check_valid(f, 1, y=2)
    assert check_valid(f, x=1, y=2)
    assert check_valid(f, 1, 2) is False
    assert check_valid(f, 1, z=2) is False
    assert check_valid(f, 1, y=1, z=2) is False
    f = make_func('x=1, *, y, z=3')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1, z=3) is incomplete
    assert check_valid(f, y=2)
    assert check_valid(f, 1, y=2)
    assert check_valid(f, x=1, y=2)
    assert check_valid(f, x=1, y=2, z=3)
    assert check_valid(f, 1, x=1, y=2) is False
    assert check_valid(f, 1, 3, y=2) is False
    f = make_func('w, x=2, *args, y, z=4')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1) is incomplete
    assert check_valid(f, 1, y=3)
    f = make_func('a, b, c=3, d=4, *args, e=5, f=6, g, h')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1) is incomplete
    assert check_valid(f, 1, 2) is incomplete
    assert check_valid(f, 1, 2, g=7) is incomplete
    assert check_valid(f, 1, 2, g=7, h=8)
    assert check_valid(f, 1, 2, 3, 4, 5, 6, 7, 8, 9) is incomplete
    f = make_func('a: int, b: float')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1) is incomplete
    assert check_valid(f, b=1) is incomplete
    assert check_valid(f, 1, 2)
    f = make_func('(a: int, b: float) -> float')
    assert check_valid(f) is incomplete
    assert check_valid(f, 1) is incomplete
    assert check_valid(f, b=1) is incomplete
    assert check_valid(f, 1, 2)
    f.__signature__ = 34
    assert check_valid(f) is False

    class RaisesValueError:

        def __call__(self) -> None:
            pass

        @property
        def __signature__(self) -> object:
            raise ValueError('Testing Python 3.4')
    f = RaisesValueError()
    assert check_valid(f) is None

def test_is_partial() -> None:
    test_is_valid(check_valid=is_partial_args, incomplete=True)
    test_is_valid_py3(check_valid=is_partial_args, incomplete=True)

def test_is_valid_curry() -> None:

    def check_curry(func, args, kwargs, incomplete=True):
        try:
            curry(func)(*args, **kwargs)
            curry(func, *args)(**kwargs)
            curry(func, **kwargs)(*args)
            curry(func, *args, **kwargs)()
            if not isinstance(func, type(lambda: None)):
                return None
            return incomplete
        except ValueError:
            return True
        except TypeError:
            return False
    check_valid = functools.partial(check_curry, incomplete=True)
    test_is_valid(check_valid=check_valid, incomplete=True)
    test_is_valid_py3(check_valid=check_valid, incomplete=True)
    check_valid = functools.partial(check_curry, incomplete=False)
    test_is_valid(check_valid=check_valid, incomplete=False)
    test_is_valid_py3(check_valid=check_valid, incomplete=False)

def test_func_keyword() -> None:

    def f(func: object = None) -> None:
        pass
    assert is_valid_args(f, (), {})
    assert is_valid_args(f, (None,), {})
    assert is_valid_args(f, (), {'func': None})
    assert is_valid_args(f, (None,), {'func': None}) is False
    assert is_partial_args(f, (), {})
    assert is_partial_args(f, (None,), {})
    assert is_partial_args(f, (), {'func': None})
    assert is_partial_args(f, (None,), {'func': None}) is False

def test_has_unknown_args() -> None:
    assert has_varargs(1) is False
    assert has_varargs(map)
    assert has_varargs(make_func('')) is False
    assert has_varargs(make_func('x, y, z')) is False
    assert has_varargs(make_func('*args'))
    assert has_varargs(make_func('**kwargs')) is False
    assert has_varargs(make_func('x, y, *args, **kwargs'))
    assert has_varargs(make_func('x, y, z=1')) is False
    assert has_varargs(make_func('x, y, z=1, **kwargs')) is False
    f = make_func('*args')
    f.__signature__ = 34
    assert has_varargs(f) is False

    class RaisesValueError:

        def __call__(self) -> None:
            pass

        @property
        def __signature__(self) -> object:
            raise ValueError('Testing Python 3.4')
    f = RaisesValueError()
    assert has_varargs(f) is None

def test_num_required_args() -> None:
    assert num_required_args(lambda: None) == 0
    assert num_required_args(lambda x: None) == 1
    assert num_required_args(lambda x, *args: None) == 1
    assert num_required_args(lambda x, **kwargs: None) == 1
    assert num_required_args(lambda x, y, *args, **kwargs: None) == 2
    assert num_required_args(map) == 2
    assert num_required_args(dict) is None

def test_has_keywords() -> None:
    assert has_keywords(lambda: None) is False
    assert has_keywords(lambda x: None) is False
    assert has_keywords(lambda x=1: None)
    assert has_keywords(lambda **kwargs: None)
    assert has_keywords(int)
    assert has_keywords(sorted)
    assert has_keywords(max)
    assert has_keywords(map) == (sys.version_info[1] >= 14)
    assert has_keywords(bytearray) is None

def test_has_varargs() -> None:
    assert has_varargs(lambda: None) is False
    assert has_varargs(lambda *args: None)
    assert has_varargs(lambda **kwargs: None) is False
    assert has_varargs(map)
    assert has_varargs(max) is None

def test_is_arity() -> None:
    assert is_arity(0, lambda: None)
    assert is_arity(1, lambda: None) is False
    assert is_arity(1, lambda x: None)
    assert is_arity(3, lambda x, y, z: None)
    assert is_arity(1, lambda x, *args: None) is False
    assert is_arity(1, lambda x, **kwargs: None) is False
    assert is_arity(1, all)
    assert is_arity(2, map) is False
    assert is_arity(2, range) is None

def test_introspect_curry_valid_py3(check_valid=is_valid_args, incomplete: bool = False) -> None:
    orig_check_valid = check_valid
    check_valid = lambda _func, *args, **kwargs: orig_check_valid(_func, args, kwargs)
    f = humpy_toolz.curry(make_func('x, y, z=0'))
    assert check_valid(f)
    assert check_valid(f, 1)
    assert check_valid(f, 1, 2)
    assert check_valid(f, 1, 2, 3)
    assert check_valid(f, 1, 2, 3, 4) is False
    assert check_valid(f, invalid_keyword=True) is False
    assert check_valid(f(1))
    assert check_valid(f(1), 2)
    assert check_valid(f(1), 2, 3)
    assert check_valid(f(1), 2, 3, 4) is False
    assert check_valid(f(1), x=2) is False
    assert check_valid(f(1), y=2)
    assert check_valid(f(x=1), 2) is False
    assert check_valid(f(x=1), y=2)
    assert check_valid(f(y=2), 1)
    assert check_valid(f(y=2), 1, z=3)
    assert check_valid(f(y=2), 1, 3) is False
    f = humpy_toolz.curry(make_func('x, y, z=0'), 1, x=1)
    assert check_valid(f) is False
    assert check_valid(f, z=3) is False
    f = humpy_toolz.curry(make_func('x, y, *args, z'))
    assert check_valid(f)
    assert check_valid(f, 0)
    assert check_valid(f(1), 0)
    assert check_valid(f(1, 2), 0)
    assert check_valid(f(1, 2, 3), 0)
    assert check_valid(f(1, 2, 3, 4), 0)
    assert check_valid(f(1, 2, 3, 4), z=4)
    assert check_valid(f(x=1))
    assert check_valid(f(x=1), 1) is False
    assert check_valid(f(x=1), y=2)

def test_introspect_curry_partial_py3() -> None:
    test_introspect_curry_valid_py3(check_valid=is_partial_args, incomplete=True)

def test_introspect_curry_py3() -> None:
    f = humpy_toolz.curry(make_func(''))
    assert num_required_args(f) == 0
    assert is_arity(0, f)
    assert has_varargs(f) is False
    assert has_keywords(f) is False
    f = humpy_toolz.curry(make_func('x'))
    assert num_required_args(f) == 0
    assert is_arity(0, f) is False
    assert is_arity(1, f) is False
    assert has_varargs(f) is False
    assert has_keywords(f)
    f = humpy_toolz.curry(make_func('x, y, z=0'))
    assert num_required_args(f) == 0
    assert is_arity(0, f) is False
    assert is_arity(1, f) is False
    assert is_arity(2, f) is False
    assert is_arity(3, f) is False
    assert has_varargs(f) is False
    assert has_keywords(f)
    f = humpy_toolz.curry(make_func('*args, **kwargs'))
    assert num_required_args(f) == 0
    assert has_varargs(f)
    assert has_keywords(f)

def test_introspect_builtin_modules() -> None:
    mods: list[ModuleType] = [builtins, functools, itertools, operator, humpy_toolz, humpy_toolz.functoolz, humpy_toolz.itertoolz, humpy_toolz.dicttoolz, humpy_toolz.recipes]
    denylist: set[object] = set()

    def add_denylist(mod: ModuleType, attr: str) -> None:
        if hasattr(mod, attr):
            denylist.add(getattr(mod, attr))
    add_denylist(builtins, 'basestring')
    add_denylist(builtins, 'NoneType')
    add_denylist(builtins, '__metaclass__')
    add_denylist(builtins, 'sequenceiterator')

    def is_missing(modname: str, name: str, func: object) -> bool:
        if name.startswith('_') and (not name.startswith('__')):
            return False
        if name.startswith('__pyx_unpickle_') or name.endswith('_cython__'):
            return False
        try:
            if issubclass(func, BaseException):
                return False
        except TypeError:
            pass
        try:
            return callable(func) and func.__module__ is not None and (modname in func.__module__) and (is_partial_args(func, (), {}) is not True) and (func not in denylist)
        except AttributeError:
            return False
    missing = {}
    for mod in mods:
        modname = mod.__name__
        for name, func in vars(mod).items():
            if is_missing(modname, name, func):
                if modname not in missing:
                    missing[modname] = []
                missing[modname].append(name)
    if missing:
        messages: list[str] = []
        for modname, names in sorted(missing.items()):
            msg = '{}:\n    {}'.format(modname, '\n    '.join(sorted(names)))
            messages.append(msg)
        message = 'Missing introspection for the following callables:\n\n'
        raise AssertionError(message + '\n\n'.join(messages))

def test_inspect_signature_property() -> None:

    class AddX:

        def __init__(self, func: Callable[..., int]) -> None:
            self.func = func

        def __call__(self, addx, *args, **kwargs):
            return addx + self.func(*args, **kwargs)

        @property
        def __signature__(self):
            sig = inspect.signature(self.func)
            params = list(sig.parameters.values())
            kind = inspect.Parameter.POSITIONAL_OR_KEYWORD
            newparam = inspect.Parameter('addx', kind)
            params = [newparam] + params
            return sig.replace(parameters=params)
    addx = AddX(lambda x: x)
    sig = inspect.signature(addx)
    assert sig == inspect.Signature(parameters=[inspect.Parameter('addx', inspect.Parameter.POSITIONAL_OR_KEYWORD), inspect.Parameter('x', inspect.Parameter.POSITIONAL_OR_KEYWORD)])
    assert num_required_args(AddX) is False
    _sigs.signatures[AddX] = (_sigs.expand_sig((0, lambda func: None)),)
    assert num_required_args(AddX) == 1
    del _sigs.signatures[AddX]

def test_inspect_wrapped_property() -> None:

    class Wrapped:

        def __init__(self, func: Callable[..., object]) -> None:
            self.func = func

        def __call__(self, *args: object, **kwargs: object) -> object:
            return self.func(*args, **kwargs)

        @property
        def __wrapped__(self) -> Callable[..., object]:
            return self.func
    func: Callable[[int], int] = lambda x: x
    wrapped = Wrapped(func)
    assert inspect.signature(func) == inspect.signature(wrapped)
    inspectbroken = True
    if sys.version_info.major > 3:
        inspectbroken = False
    if sys.version_info.minor == 11 and sys.version_info.micro > 8:
        inspectbroken = False
    if sys.version_info.minor == 12 and sys.version_info.micro > 2:
        inspectbroken = False
    if sys.version_info.minor > 12:
        inspectbroken = False
    if inspectbroken:
        assert num_required_args(Wrapped) is None
        _sigs.signatures[Wrapped] = (_sigs.expand_sig((0, lambda func: None)),)
    assert num_required_args(Wrapped) == 1



from .dicttoolz import *
from .functoolz import *
from .itertoolz import *
from .recipes import *
from functools import partial, reduce

__version__ = '1.1.0'
sorted = sorted
map = map
filter = filter
comp = compose
flip = functoolz.flip = curry(functoolz.flip)
memoize = functoolz.memoize = curry(functoolz.memoize)
from . import curried

functoolz._sigs.update_signature_registry()
__toolz_version__ = '1.1.0'

def __getattr__(name):
    if name == '__version__':
        from importlib.metadata import version
        rv = version('humpy_cytoolz')
        globals()[name] = rv
        return rv
    raise AttributeError(f'module {__name__!r} has no attribute {name!r}')



"""I type, you type, we all `type` for `theTypes`."""
from hunterMakesPy.theTypesCallableFunction import CallableFunction as CallableFunction
from typing import Protocol, Self

type identifierDotAttribute = str
"""`str` (***str***ing) representing a dotted attribute identifier.

`TypeAlias` for a `str` `object` using dot notation to access an attribute, such as 'scipy.signal.windows'.
"""

class Ordinals(Protocol):
	"""Any Python `object` `type` that may be ordered before or after a comparable `object` `type` by comparison operators."""

	def __le__(self: Self, not_self_selfButSelfSelf_youKnow: Self, /) -> bool:
		"""Comparison by "***l***ess than or ***e***qual to"."""
		...

	def __lt__(self: Self, otherSelfWhichIsNotAnOxymoron: Self, /) -> bool:
		"""Comparison by "***l***ess ***t***han"."""
		...



"""Manipulate data structures with encoding, extraction, and merging utilities.

(AI generated docstring)

You can use this module to transform NumPy arrays [1] into compact run-length encoded
strings, extract strings from arbitrarily nested data structures, and merge multiple
dictionaries with list values. The module provides specialized utilities for working with
Cartesian mappings, heterogeneous nested data, and dictionary consolidation operations.

The run-length encoding function produces self-decoding string representations optimized for
large arrays with repetitive patterns. The string extraction function recursively traverses
nested structures to collect all convertible string values. The dictionary merging function
consolidates multiple dictionaries while offering optional deduplication and sorting.

Contents
--------
Functions
	autoDecodingRLE
		Transform a NumPy array into a compact, self-decoding run-length encoded string representation.
	stringItUp
		Convert every element in input data structures to strings.
	updateExtendPolishDictionaryLists
		Merge multiple dictionaries with list values into a single dictionary.

References
----------
[1] NumPy - Context7
	https://numpy.org/doc/stable/reference/index.html

"""
from charset_normalizer import CharsetMatch
from collections.abc import Mapping
from humpy_tlz.functoolz import identity
from humpy_tlz.recipes import partitionby
from hunterMakesPy import Ordinals
from numpy import integer
from numpy.typing import NDArray
from typing import Any, cast
import charset_normalizer
import sys

# TODO refine autoDecodingRLE.
def autoDecodingRLE(arrayTarget: NDArray[integer[Any]], *, assumeAddSpaces: bool = False) -> str:
	"""Transform a NumPy array into a compact, self-decoding run-length encoded string representation.

	This function converts a NumPy array into a string that, when evaluated as Python code, recreates
	the original array structure. The function employs two compression strategies:
	1. Python's `range` syntax for consecutive integer sequences
	2. Multiplication syntax for repeated elements

	The resulting string representation is designed to be both human-readable and space-efficient,
	especially for large cartesian mappings with repetitive patterns. When this string is used as a
	data source, Python will automatically decode it into Python `list`, which if used as an argument
	to `numpy.array()`, will recreate the original array structure.

	Parameters
	----------
	arrayTarget : NDArray[integer[Any]]
		The NumPy array to be encoded.
	assumeAddSpaces : bool = False
		Affects internal length comparison during compression decisions. This parameter doesn't
		directly change output format but influences whether `range` or multiplication syntax is
		preferred in certain cases. The parameter exists because the Abstract Syntax Tree (AST)
		inserts spaces in its string representation.

	Returns
	-------
	encodedString : str
		A string representation of the array using run-length encoding that, when evaluated as Python
		code, reproduces the original array structure.

	Notes
	-----
	The "autoDecoding" feature means that the string representation evaluates directly to the desired
	data structure without explicit decompression steps.

	The encoded string uses only builtins — no imports are needed to decode it.
	"""
	def measurerCalculatesAdjustedLength(string: str) -> int:
		"""`assumeAddSpaces` characters: `,` 1; `]*` 2."""
		return len(string) + assumeAddSpaces * (string.count(',') + string.count(']*') * 2)

	# TODO there is only ONE call to this. Refactor to inline it and remove the function definition.
	def encoderConvertsTokenToString(token: int | tuple[int, int]) -> str:
		string: str = str(token)
		if isinstance(token, tuple):
			rangeStart, rangeEnd = token
			if rangeStart == 0:
				string = f"*range({rangeEnd})"
			else:
				string = f"*range({rangeStart},{rangeEnd})"
		return string

	def encoderTransformsOneDimensionalSequenceToString(listIntegers: list[int]) -> str:
		stringRepresentingArray: str = '[]'

		if listIntegers:
			listTokens: list[int | tuple[int, int]] = []
			index: int = 0
			countIntegers: int = len(listIntegers)

			while index < countIntegers:
				countConsecutive: int = 1
				while (index + countConsecutive < countIntegers) and (listIntegers[index + countConsecutive] == listIntegers[index] + countConsecutive):
					countConsecutive += 1

				integerStart: int = listIntegers[index]
				integerEnd: int = integerStart + countConsecutive
				stringRangeSyntax: str = ''
				if integerStart == 0:
					stringRangeSyntax = f"*range({integerEnd})"
				else:
					stringRangeSyntax = f"*range({integerStart},{integerEnd})"
				stringCommaSeparated: str = ','.join(str(integerStart + offset) for offset in range(countConsecutive))

				if measurerCalculatesAdjustedLength(stringRangeSyntax) < measurerCalculatesAdjustedLength(stringCommaSeparated):
					listTokens.append((integerStart, integerEnd))
				else:
					for indexOffset in range(countConsecutive):
						listTokens.append(integerStart + indexOffset)  # noqa: PERF401
				index += countConsecutive

			listSegments: list[str] = []
			listBuffer: list[str] = []

			for groupTokens in partitionby(identity, listTokens):
				listGroup: list[int | tuple[int, int]] = list(groupTokens)
				tokenActive: int | tuple[int, int] = listGroup[0]
				countRepetitions: int = len(listGroup)

				stringTokenActive: str = encoderConvertsTokenToString(tokenActive)

				if 1 < countRepetitions:
					stringMultiplicationSyntax: str = f"[{stringTokenActive}]*{countRepetitions}"
					stringListSyntax: str = "[" + ",".join([stringTokenActive] * countRepetitions) + "]"
					if measurerCalculatesAdjustedLength(stringMultiplicationSyntax) < measurerCalculatesAdjustedLength(stringListSyntax):
						if listBuffer:
							listSegments.append('[' + ','.join(listBuffer) + ']')
							listBuffer = []
						listSegments.append(stringMultiplicationSyntax)
						continue

				listBuffer.extend([stringTokenActive] * countRepetitions)

			if listBuffer:
				listSegments.append('[' + ','.join(listBuffer) + ']')

			if listSegments:
				stringRepresentingArray = '+'.join(listSegments)

		return stringRepresentingArray

	def encoderTransformsArrayRecursively(arraySlice: NDArray[integer[Any]]) -> str:
		if arraySlice.ndim == 0:
			return str(int(arraySlice))
		if arraySlice.ndim == 1:
			return encoderTransformsOneDimensionalSequenceToString(arraySlice.tolist())
		return '[' + ','.join(encoderTransformsArrayRecursively(arraySlice[indexRow]) for indexRow in range(arraySlice.shape[0])) + ']'

	return encoderTransformsArrayRecursively(arrayTarget)

def stringItUp(*scrapPile: Any) -> list[str]:
	"""Convert, if possible, every element in the input data structure to a string.

	Order is not preserved or readily predictable.

	Parameters
	----------
	*scrapPile : Any
		(scrap2pile) One or more data structures to unpack and convert to strings.

	Returns
	-------
	listStrungUp : list[str]
		(list2strung2up) A `list` of string versions of all convertible elements.

	"""
	scrap: Any = None
	listStrungUp: list[str] = []

	def drill(KitKat: Any) -> None:
		if isinstance(KitKat, str):
			listStrungUp.append(KitKat)
		elif (KitKat is None) or (isinstance(KitKat, (bool, bytearray, bytes, complex, float, int))):
			listStrungUp.append(str(KitKat))
		elif isinstance(KitKat, memoryview):
			decodedString: CharsetMatch | None = charset_normalizer.from_bytes(KitKat.tobytes()).best()
			if decodedString:
				listStrungUp.append(str(decodedString))
		elif isinstance(KitKat, dict):
			DictDact: dict[Any, Any] = cast(dict[Any, Any], KitKat)
			for broken, piece in DictDact.items():
				drill(broken)
				drill(piece)
		elif isinstance(KitKat, (list, tuple, set, frozenset, range)):
			for kit in KitKat: # pyright: ignore[reportUnknownVariableType]
				drill(kit)
		elif hasattr(KitKat, '__iter__'):  # Unpack other iterables
			for kat in KitKat:
				drill(kat)
		else:
			try:
				sharingIsCaring: str = KitKat.__str__()
				listStrungUp.append(sharingIsCaring)
			except AttributeError:
				pass
			except TypeError:  # "The error traceback provided indicates that there is an issue when calling the __str__ method on an object that does not have this method properly defined, leading to a TypeError."
				pass
			except:
				message: str = (f"\nWoah! I received '{repr(KitKat)}'.\nTheir report card says, 'Plays well with others: Needs improvement.'\n")
				sys.stderr.write(message)
				raise
	try:
		for scrap in scrapPile:
			drill(scrap)
	except RecursionError:
		listStrungUp.append(repr(scrap))
	return listStrungUp

def updateExtendPolishDictionaryLists[小于: Ordinals](*dictionaryLists: Mapping[str, list[小于] | set[小于] | tuple[小于, ...]], destroyDuplicates: bool = False, reorderLists: bool = False, killErroneousDataTypes: bool = False) -> dict[str, list[小于]]:
	"""Merge multiple dictionaries with `list` values into a single dictionary with the `list` values merged.

	Plus options to destroy duplicates, sort `list` values, and handle erroneous data types.

	Parameters
	----------
	*dictionaryLists : Mapping[str, list[Any] | set[Any] | tuple[Any, ...]]
		Variable number of dictionaries to be merged. If only one dictionary is passed, it will be
		"polished".
	destroyDuplicates : bool = False
		If `True`, removes duplicate elements from the `list`. Defaults to `False`.
	reorderLists : bool = False
		If `True`, sorts each `list` value. Defaults to `False`. The elements must be comparable;
		otherwise, a `TypeError` will be raised.
	killErroneousDataTypes : bool = False
		If `True`, suppresses any `TypeError` `Exception` and omits the dictionary key or value that
		caused the `Exception`. Defaults to `False`.

	Returns
	-------
	ePluribusUnum : dict[str, list[Any]]
		A single dictionary with merged and optionally "polished" `list` values.

	Notes
	-----
	The returned value, `ePluribusUnum`, is a so-called primitive dictionary (`dict`). Furthermore,
	every dictionary key is a so-called primitive string (*cf.* `str()`) and every dictionary value
	is a so-called primitive `list` (`list`). If `dictionaryLists` has other data types, the data
	types will not be preserved. That could have unexpected consequences. Conversion from the
	original data type to a `list`, for example, may not preserve the order even if you want the
	order to be preserved.
	"""
	ePluribusUnum: dict[str, list[小于]] = {}

	for dictionaryListTarget in dictionaryLists:
		for keyName, keyValue in dictionaryListTarget.items():
			try:
				ImaStr = str(keyName)
				ImaList: list[小于] = list(keyValue)
				ePluribusUnum.setdefault(ImaStr, []).extend(ImaList)
			except TypeError:
				if killErroneousDataTypes:
					continue
				else:
					raise

	if destroyDuplicates:
		for ImaStr, ImaList in ePluribusUnum.items():
			ePluribusUnum[ImaStr] = list(dict.fromkeys(ImaList))
	if reorderLists:
		for ImaStr, ImaRichComparisonSupporter in ePluribusUnum.items():
			ePluribusUnum[ImaStr] = sorted(ImaRichComparisonSupporter)

	return ePluribusUnum



from .dicttoolz import *
from .functoolz import *
from .itertoolz import *
from .recipes import *
from functools import partial, reduce

__version__ = '1.1.0'
sorted = sorted
map = map
filter = filter
comp = compose
from . import curried, sandbox

functoolz._sigs.create_signature_registry()

def __getattr__(name):
    if name == '__version__':
        from importlib.metadata import version
        rv = version('humpy_toolz')
        globals()[name] = rv
        return rv
    raise AttributeError(f'module {__name__!r} has no attribute {name!r}')



"""
Alternate namespace for toolz such that all functions are curried

Currying provides implicit partial evaluation of all functions

Example:

    Get usually requires two arguments, an index and a collection
    >>> from humpy_toolz.curried import get
    >>> get(0, ('a', 'b'))
    'a'

    When we use it in higher order functions we often want to pass a partially
    evaluated form
    >>> data = [(1, 2), (11, 22), (111, 222)]
    >>> list(map(lambda seq: get(0, seq), data))
    [1, 11, 111]

    The curried version allows simple expression of partial evaluation
    >>> list(map(get(0), data))
    [1, 11, 111]

See Also:
    humpy_toolz.functoolz.curry
"""
from . import operator
from .exceptions import merge, merge_with
from humpy_toolz import (
	apply, comp, complement, compose, compose_left, concat, concatv, count, curry, diff, first, flip, frequencies,
	identity, interleave, isdistinct, isiterable, juxt, last, memoize, merge_sorted, peek, pipe, second, thread_first,
	thread_last)
import humpy_toolz

accumulate = humpy_toolz.curry(humpy_toolz.accumulate)
assoc = humpy_toolz.curry(humpy_toolz.assoc)
assoc_in = humpy_toolz.curry(humpy_toolz.assoc_in)
cons = humpy_toolz.curry(humpy_toolz.cons)
countby = humpy_toolz.curry(humpy_toolz.countby)
dissoc = humpy_toolz.curry(humpy_toolz.dissoc)
do = humpy_toolz.curry(humpy_toolz.do)
drop = humpy_toolz.curry(humpy_toolz.drop)
excepts = humpy_toolz.curry(humpy_toolz.excepts)
filter = humpy_toolz.curry(humpy_toolz.filter)
get = humpy_toolz.curry(humpy_toolz.get)
get_in = humpy_toolz.curry(humpy_toolz.get_in)
groupby = humpy_toolz.curry(humpy_toolz.groupby)
interpose = humpy_toolz.curry(humpy_toolz.interpose)
itemfilter = humpy_toolz.curry(humpy_toolz.itemfilter)
itemmap = humpy_toolz.curry(humpy_toolz.itemmap)
iterate = humpy_toolz.curry(humpy_toolz.iterate)
join = humpy_toolz.curry(humpy_toolz.join)
keyfilter = humpy_toolz.curry(humpy_toolz.keyfilter)
keymap = humpy_toolz.curry(humpy_toolz.keymap)
map = humpy_toolz.curry(humpy_toolz.map)
mapcat = humpy_toolz.curry(humpy_toolz.mapcat)
nth = humpy_toolz.curry(humpy_toolz.nth)
partial = humpy_toolz.curry(humpy_toolz.partial)
partition = humpy_toolz.curry(humpy_toolz.partition)
partition_all = humpy_toolz.curry(humpy_toolz.partition_all)
partitionby = humpy_toolz.curry(humpy_toolz.partitionby)
peekn = humpy_toolz.curry(humpy_toolz.peekn)
pluck = humpy_toolz.curry(humpy_toolz.pluck)
random_sample = humpy_toolz.curry(humpy_toolz.random_sample)
reduce = humpy_toolz.curry(humpy_toolz.reduce)
reduceby = humpy_toolz.curry(humpy_toolz.reduceby)
remove = humpy_toolz.curry(humpy_toolz.remove)
sliding_window = humpy_toolz.curry(humpy_toolz.sliding_window)
sorted = humpy_toolz.curry(humpy_toolz.sorted)
tail = humpy_toolz.curry(humpy_toolz.tail)
take = humpy_toolz.curry(humpy_toolz.take)
take_nth = humpy_toolz.curry(humpy_toolz.take_nth)
topk = humpy_toolz.curry(humpy_toolz.topk)
unique = humpy_toolz.curry(humpy_toolz.unique)
update_in = humpy_toolz.curry(humpy_toolz.update_in)
valfilter = humpy_toolz.curry(humpy_toolz.valfilter)
valmap = humpy_toolz.curry(humpy_toolz.valmap)
del exceptions
del humpy_toolz



from .dev_skip_test import dev_skip_test
from humpy_cytoolz import curry, identity, keyfilter, merge_with, valfilter
from humpy_cytoolz.utils import raises
import difflib
import humpy_cytoolz

skipped_doctests = ['get_in']

@curry
def isfrommod(modname, func):
    mod = getattr(func, '__module__', '') or ''
    return mod.startswith(modname) or 'toolz.functoolz.curry' in str(type(func))

def convertdoc(doc):
    """ Convert docstring from `humpy_toolz` to `humpy_cytoolz`."""
    if hasattr(doc, '__doc__'):
        doc = doc.__doc__
    doc = doc.replace('humpy_toolz', 'humpy_cytoolz')
    doc = doc.replace('dictcytoolz', 'dicttoolz')
    doc = doc.replace('funccytoolz', 'functoolz')
    doc = doc.replace('itercytoolz', 'itertoolz')
    doc = doc.replace('cytoolz.readthedocs', 'toolz.readthedocs')
    return doc

@dev_skip_test
def test_docstrings_uptodate():
    import humpy_toolz
    differ = difflib.Differ()
    toolz_dict = valfilter(isfrommod('humpy_toolz'), humpy_toolz.__dict__)
    cytoolz_dict = valfilter(isfrommod('humpy_cytoolz'), humpy_cytoolz.__dict__)
    toolz_dict = valfilter(lambda x: getattr(x, '__doc__', ''), toolz_dict)
    toolz_dict = keyfilter(lambda x: x in cytoolz_dict, toolz_dict)
    cytoolz_dict = keyfilter(lambda x: x in toolz_dict, cytoolz_dict)
    d = merge_with(identity, toolz_dict, cytoolz_dict)
    for key, (toolz_func, cytoolz_func) in d.items():
        toolz_doc = '\n'.join((line.strip() for line in convertdoc(toolz_func).splitlines()))
        cytoolz_doc = '\n'.join((line.strip() for line in cytoolz_func.__doc__.splitlines()))
        if toolz_doc not in cytoolz_doc:
            diff = list(differ.compare(toolz_doc.splitlines(), cytoolz_doc.splitlines()))
            fulldiff = list(diff)
            while diff and diff[0].startswith('+'):
                diff.pop(0)
            while diff and diff[-1].startswith('+'):
                diff.pop()

            def checkbad(line):
                return line.startswith('+') and (not ('# doctest: +SKIP' in line and key in skipped_doctests))
            if any(map(checkbad, diff)):
                assert False, 'Error: cytoolz.{} has a bad docstring:\n{}\n'.format(key, '\n'.join(fulldiff))

def test_get_in_doctest():
    raises(KeyError, lambda: humpy_cytoolz.get_in(['y'], {}, no_default=True))



"""Perform file system operations and dynamic module imports.

(AI generated docstring)

You can use this module to import identifiers from modules by logical path or file path,
safely create directories, format and write Python source code, and write strings to files
or text streams. The module provides utilities for dynamic code loading, safe file operations,
and automated Python source code formatting using autoflake [1] and isort [2].

The import functions support both logical module paths using dot notation and direct file
paths for loading Python modules. The directory creation function suppresses errors when
directories already exist. The Python writing function automatically formats source code
before writing, removing unused imports and sorting import statements according to
configurable settings.

Contents
--------
Functions
	importLogicalPath2Identifier
		Import an identifier from a module using its logical path.
	importPathFilename2Identifier
		Load an identifier from a Python file.
	makeDirsSafely
		Create parent directories for a given path safely.
	writePython
		Format and write Python source code to a file or text stream.
	writeStringToHere
		Write a string to a file or text stream.

Module Constants
	settings_autoflakeDEFAULT
		Default configuration for autoflake formatter.
	settings_isortDEFAULT
		Default configuration for isort formatter.

References
----------
[1] autoflake - Context7
	https://github.com/PyCQA/autoflake
[2] isort - Context7
	https://pycqa.github.io/isort/

"""
from autoflake import fix_code as autoflake_fix_code
from hunterMakesPy import identifierDotAttribute
from isort import code as isort_code
from os import PathLike
from pathlib import Path, PurePath
from typing import Any, overload, TYPE_CHECKING
import contextlib
import importlib
import importlib.util
import io

if TYPE_CHECKING:
	from importlib.machinery import ModuleSpec
	from types import ModuleType

def importLogicalPath2Identifier[归个](logicalPathModule: identifierDotAttribute, identifier: str, packageIdentifierIfRelative: str | None = None) -> 归个:
	"""Import an `identifier`, such as a function or `class`, from a module using its logical path.

	This function imports a module and retrieves a specific attribute (function, class, or other object) from that module.

	Parameters
	----------
	logicalPathModule : identifierDotAttribute
		The logical path to the module, using dot notation (e.g., 'scipy.signal.windows').
	identifier : str
		The identifier of the object to retrieve from the module.
	packageIdentifierIfRelative : str | None = None
		The package name to use as the anchor point if `logicalPathModule` is a relative import. `None` means an absolute import.

	Returns
	-------
	identifierImported : 归个
		The identifier (function, class, or object) retrieved from the module.

	"""
	moduleImported: ModuleType = importlib.import_module(logicalPathModule, packageIdentifierIfRelative)
	return getattr(moduleImported, identifier)

def importPathFilename2Identifier[归个](pathFilename: PathLike[Any] | PurePath, identifier: str, moduleIdentifier: str | None = None) -> 归个:
	"""Load an identifier from a Python file.

	This function imports a specified Python file as a module, extracts an identifier from it by name, and returns that
	identifier.

	Parameters
	----------
	pathFilename : PathLike[Any] | PurePath
		Path to the Python file to import.
	identifier : str
		Name of the identifier to extract from the imported module.
	moduleIdentifier : str | None = None
		Name to use for the imported module. If `None`, the filename stem is used.

	Returns
	-------
	identifierImported : 归个
		The identifier extracted from the imported module.

	Raises
	------
	ImportError
		If the file cannot be imported or the importlib specification is invalid.
	AttributeError
		If the identifier does not exist in the imported module.

	"""
	pathFilename = Path(pathFilename)

	importlibSpecification: ModuleSpec | None = importlib.util.spec_from_file_location(moduleIdentifier or pathFilename.stem, pathFilename)
	if importlibSpecification is None or importlibSpecification.loader is None:
		message: str = f"I received\n\t`{pathFilename = }`,\n\t`{identifier = }`, and\n\t`{moduleIdentifier = }`.\n\tAfter loading, \n\t`importlibSpecification` {'is `None`' if importlibSpecification is None else 'has a value'} and\n\t`importlibSpecification.loader` is unknown."
		raise ImportError(message)

	moduleImported_jk_hahaha: ModuleType = importlib.util.module_from_spec(importlibSpecification)
	importlibSpecification.loader.exec_module(moduleImported_jk_hahaha)
	return getattr(moduleImported_jk_hahaha, identifier)

# SEMIOTICS "Dirs"
def makeDirsSafely(pathFilename: Any) -> None:
	"""Create parent directories for a given path safely.

	This function attempts to create all necessary parent directories for a given path. If the directory already exists or if
	there's an `OSError` during creation, it will silently continue without raising an exception.

	Parameters
	----------
	pathFilename : Any
		A path-like object or file object representing the path for which to create parent directories. If it's an IO stream
		object, no directories will be created.

	"""
	if not isinstance(pathFilename, io.IOBase):
		with contextlib.suppress(OSError):
			Path(pathFilename).parent.mkdir(parents=True, exist_ok=True)

settings_autoflakeDEFAULT: dict[str, list[str] | bool] = {
	'additional_imports': [],
	'expand_star_imports': True,
	'remove_all_unused_imports': True,
	'remove_duplicate_keys': False,
	'remove_unused_variables': False,
}

settings_isortDEFAULT: dict[str, bool | int | str | list[str]] = {
	"combine_as_imports": True,
	"force_alphabetical_sort_within_sections": True,
	"from_first": True,
	"honor_noqa": True,
	"indent": "\t",
	"line_length": 140,
	"lines_after_imports": 1,
	"lines_between_types": 0,
	"multi_line_output": 4,
	"no_sections": True,
	"use_parentheses": True,
}

@overload
def writePython(pythonSource: str, pathFilename: PathLike[Any] | PurePath, settings: dict[str, dict[str, Any]] | None = None) -> Path: ...
@overload
def writePython(pythonSource: str, pathFilename: io.TextIOBase, settings: dict[str, dict[str, Any]] | None = None) ->  io.TextIOBase: ...
def writePython(pythonSource: str, pathFilename: PathLike[Any] | PurePath | io.TextIOBase, settings: dict[str, dict[str, Any]] | None = None) -> Path | io.TextIOBase:
	"""Format and write Python source code to a file or text stream.

	(AI generated docstring)

	This function processes Python source code through autoflake and isort formatters before writing to the specified destination.
	The formatters remove unused imports, sort imports, and apply consistent code style according to the provided or default
	settings.

	Parameters
	----------
	pythonSource : str
		The Python source code to format and write.
	pathFilename : PathLike[Any] | PurePath | io.TextIOBase
		The target destination: either a file path or an open text stream.
	settings : dict[str, dict[str, Any]] | None = None
		Configuration for the formatters. Keys are `'autoflake'` and `'isort'`, each mapping to a dictionary of formatter-specific
		settings. If `None`, default settings are used for both formatters.

	"""
	if settings is None:
		settings = {}
# TODO The formatting system needs a complete rethink.
	settings_autoflake: dict[str, Any] = settings.get('autoflake', settings_autoflakeDEFAULT)
	pythonSource = autoflake_fix_code(pythonSource, **settings_autoflake)

	settings_isort: dict[str, Any] = settings.get('isort', settings_isortDEFAULT)
	pythonSource = isort_code(pythonSource, **settings_isort)
	return writeStringToHere(pythonSource + '\n', pathFilename)

@overload
def writeStringToHere(this: str, pathFilename: PathLike[Any] | PurePath) -> Path: ...
@overload
def writeStringToHere(this: str, pathFilename: io.TextIOBase) -> io.TextIOBase: ...
def writeStringToHere(this: str, pathFilename: PathLike[Any] | PurePath | io.TextIOBase) -> Path | io.TextIOBase:
	"""Write a string to a file or text stream.

	This function writes a string to either a file path or an open text stream. For file paths, it creates parent directories as
	needed and writes with UTF-8 encoding. For text streams, it writes directly to the stream and flushes the buffer.

	Parameters
	----------
	this : str
		The string content to write.
	pathFilename : PathLike[Any] | PurePath | io.TextIOBase
		The target destination: either a file path or an open text stream.

	"""
	if isinstance(pathFilename, io.TextIOBase):
		pathFilename.write(str(this))
		pathFilename.flush()
	else:
		pathFilename = Path(pathFilename)
		makeDirsSafely(pathFilename)
		pathFilename.write_text(str(this), encoding='utf-8')
	return pathFilename



from typing import Protocol
import types

class _AnyCallable[ReturnT](Protocol):
	def __call__(self, *args: object, **kwargs: object) -> ReturnT: ...

type Signature = _AnyCallable[object]
type SignatureInfo = dict[str, list[Signature]]

cytoolz_info: dict[str, SignatureInfo]
module_info: dict[types.ModuleType | str, SignatureInfo]

def create_signature_registry(module_info: dict[types.ModuleType | str, SignatureInfo]) -> None: ...

def update_signature_registry() -> None: ...

def _is_arity(func: Signature, n: int) -> bool: ...

def _has_varargs(func: Signature) -> bool: ...

def _has_keywords(func: Signature) -> bool: ...

def _num_required_args(func: Signature) -> int: ...

def _is_partial_args(func: Signature, args: tuple[object, ...], kwargs: dict[str, object]) -> bool: ...

def _is_valid_args(func: Signature, args: tuple[object, ...], kwargs: dict[str, object]) -> bool: ...



from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from typing import Any, overload

__all__ = ("countby", "partitionby")

@overload
def countby[T, K](key: Callable[[T], K], seq: Iterable[T]) -> dict[K, int]: ...
@overload
def countby[V](key: Any, seq: Iterable[Sequence[V] | Mapping[Any, V]]) -> dict[V, int]: ...
def countby[T, K, V]( key: Callable[[T], K] | Any, seq: Iterable[T] | Iterable[Sequence[V] | Mapping[Any, V]] ) -> dict[K, int] | dict[V, int]: ...

class partitionby[T](Iterator[tuple[T, ...]]):
	def __init__(self, func: Callable[[T], Any], seq: Iterable[T]) -> None: ...
	def __iter__(self) -> partitionby[T]: ...
	def __next__(self) -> tuple[T, ...]: ...



import doctest
import humpy_toolz

def test_doctests() -> None:
    humpy_toolz.__test__ = {}
    for name, func in vars(humpy_toolz).items():
        if isinstance(func, humpy_toolz.curry):
            humpy_toolz.__test__[name] = func.func
    assert doctest.testmod(humpy_toolz).failed == 0
    del humpy_toolz.__test__



"""Validate parameters and parse input with defensive error handling.

(AI generated docstring)

You can use this module to validate and convert input sequences to integers, determine
concurrency limits based on flexible parameter specifications, and interpret string values
as boolean or None types. The module provides strict validation functions that follow
fail-early principles with descriptive error messages for debugging.

The integer validation function accepts various numeric types including strings, floats,
complex numbers, and binary data, converting them to integers while detecting ambiguous
or incompatible values. The concurrency limit function interprets boolean, integer, and
float values to compute processor counts with support for absolute limits, fractional
allocation, and reserved-processor specifications. The string interpretation function
attempts to parse string values as True, False, or None to handle parameter type mismatches
gracefully.

Contents
--------
Functions
	defineConcurrencyLimit
		Determine the concurrency limit based on flexible parameter specifications.
	intInnit
		Validate and convert input values to a list of integers.
	oopsieKwargsie
		Interpret a string as True, False, or None to avoid exceptions.

References
----------
[1] multiprocessing - Context7
	https://docs.python.org/3/library/multiprocessing.html
[2] charset-normalizer - Context7
	https://github.com/Ousret/charset_normalizer

"""
from collections.abc import Iterable, Sized
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING
import charset_normalizer
import multiprocessing

if TYPE_CHECKING:
	from charset_normalizer.models import CharsetMatch

@dataclass
class ErrorMessageContext:
	"""Context information for constructing error messages.

	Parameters
	----------
	parameterValue : Any = None
		The value that caused the error.
	parameterValueType : str | None = None
		The name of the type of the parameter value.
	containerType : str | None = None
		The name of the type of the container holding the parameter.
	isElement : bool = False
		Whether the parameter is an element within a container.

	"""

	parameterValue: Any = None
	parameterValueType: str | None = None
	containerType: str | None = None
	isElement: bool = False

def _constructErrorMessage(context: ErrorMessageContext, parameterName: str, parameterType: type[Any] | None) -> str:
	"""Construct error message from available context using template.

	I received ["value" | a value | `None`] [of type `type` | `None`] [as an element in | `None`] [a `containerType` type |
	`None`] but `parameterName` must have integers [in type(s) `parameterType` | `None`].

	Hypothetically, this is a prototype that can be generalized to other functions. In this package and a few of my other
	packages, I have developed standardized error messages, but those are quite different from this. I will certainly continue to
	develop this kind of functionality, and this function will influence things.

	Parameters
	----------
	context : ErrorMessageContext
		The error context containing parameter value, type, and container information.
	parameterName : str
		The name of the parameter that caused the error.
	parameterType : type[Any] | None
		The expected type of the parameter, used in error messages.

	Returns
	-------
	errorMessage : str
		The constructed error message string.

	"""
	messageParts: list[str] = ["I received "]

	if context.parameterValue is not None and not isinstance(context.parameterValue, (bytes, bytearray, memoryview)):
		messageParts.append(f'"{context.parameterValue}"')
	else:
		messageParts.append("a value")

	if context.parameterValueType:
		messageParts.append(f" of type `{context.parameterValueType}`")

	if context.isElement:
		messageParts.append(" as an element in")

	if context.containerType:
		messageParts.append(f" a `{context.containerType}` type")

	messageParts.append(f" but {parameterName} must have integers")

	if parameterType:
		messageParts.append(f" in type(s) `{parameterType}`")

	return "".join(messageParts)

def defineConcurrencyLimit(*, limit: bool | float | int | None, cpuTotal: int = multiprocessing.cpu_count()) -> int:
	"""Determine the concurrency limit based on the provided parameter.

	Tests for this function can be run with:
	`from hunterMakesPy.tests.test_parseParameters import PytestFor_defineConcurrencyLimit`

	Parameters
	----------
	limit : bool | float | int | None
		Whether and how to limit CPU usage. See notes and examples for details how to describe the options to your users.
	cpuTotal : int = multiprocessing.cpu_count()
		The total number of CPUs available in the system. Default is `multiprocessing.cpu_count()`.

	Returns
	-------
	concurrencyLimit : int
		The calculated concurrency limit, ensuring it is at least 1.

	Notes
	-----
	Consider using `hunterMakesPy.oopsieKwargsie()` to handle malformed inputs. For example:

	```
	if not (CPUlimit is None or isinstance(CPUlimit, (bool, int, float))):
		CPUlimit = oopsieKwargsie(CPUlimit)
	```

	Example parameters
	------------------
	```python
	CPUlimit: bool | float | int | None
	CPUlimit: bool | float | int | None = None
	```

	Example docstring
	-----------------
	```python

	Arguments
	---------
	CPUlimit: bool | float | int | None
		Whether and how to limit the the number of available processors used by the function. See notes for details.

	Notes
	-----
	Limits on CPU usage, `CPUlimit`:
		- `False`, `None`, or `0`: No limits on processor usage; uses all available processors. All other values will potentially limit processor usage.
		- `True`: Yes, limit the processor usage; limits to 1 processor.
		- `int >= 1`: The maximum number of available processors to use.
		- `0 < float < 1`: The maximum number of processors to use expressed as a fraction of available processors.
		- `-1 < float < 0`: The number of processors to *not* use expressed as a fraction of available processors.
		- `int <= -1`: The number of available processors to *not* use.
		- If the value of `CPUlimit` is a `float` greater than 1 or less than -1, the function truncates the value to an `int` with the same sign as the `float`.
	```

	"""
	concurrencyLimit: int = cpuTotal

	if isinstance(limit, str):
		limitFromString: bool | None | str = oopsieKwargsie(limit)
		if isinstance(limitFromString, str):
			try:
				limit = float(limitFromString)
			except ValueError as ERRORmessage:
				message: str = f"I received '{limitFromString}', but it must be a number, `True`, `False`, or `None`."
				raise ValueError(message) from ERRORmessage
		else:
			limit = limitFromString
	if isinstance(limit, float) and 1 <= abs(limit):
		limit = round(limit)
	if limit is None or limit is False or limit == 0:
		pass
	elif limit is True:
		concurrencyLimit = 1
	elif 1 <= limit:
		concurrencyLimit = int(limit)
	elif 0 < limit < 1:
		concurrencyLimit = round(limit * cpuTotal)
	elif -1 < limit < 0:
		concurrencyLimit = cpuTotal - abs(round(limit * cpuTotal))
	elif limit <= -1:
		concurrencyLimit = cpuTotal - abs(int(limit))

	return max(int(concurrencyLimit), 1)

# ruff: noqa: TRY301

def intInnit(listInt_Allegedly: Iterable[Any], parameterName: str | None = None, parameterType: type[Any] | None = None) -> list[int]:
	"""Validate and convert input values to a `list` of integers.

	Accepts various numeric types and attempts to convert them into integers while providing descriptive error messages. This
	package includes Pytest tests that can be imported and run: `from hunterMakesPy.tests.test_parseParameters import
	PytestFor_intInnit`.

	Parameters
	----------
	listInt_Allegedly : Iterable[Any]
		The input sequence that should contain integer-compatible values. Accepts integers, strings, floats, complex numbers, and
		binary data. Rejects boolean values and non-integer numeric values.
	parameterName : str | None = None
		Name of the parameter from your function for which this function is validating the input validated. If there is an error
		message, it provides context to your user. Defaults to 'the parameter'.
	parameterType : type[Any] | None = None
		Expected type(s) of the parameter, used in error messages.

	Returns
	-------
	listValidated : list[int]
		A `list` containing validated integers.

	Raises
	------
	ValueError
		When the input is empty or contains non-integer compatible values.
	TypeError
		When an element is a boolean or incompatible type.
	RuntimeError
		If the input sequence length changes during iteration.

	Notes
	-----
	The function performs strict validation and follows fail-early principles to catch potential issues before they become catastrophic.

	"""
	parameterName = parameterName or 'the parameter'
	parameterType = parameterType or list

	if not listInt_Allegedly:
		message: str = f"I did not receive a value for {parameterName}, but it is required."
		raise ValueError(message)

	# Be nice, and assume the input container is valid and every element is valid.
	# Nevertheless, this is a "fail-early" step, so reject ambiguity and try to induce errors now that could be catastrophic later.
	try:
		iter(listInt_Allegedly)
		lengthInitial: int | None = None
		if isinstance(listInt_Allegedly, Sized):
			lengthInitial = len(listInt_Allegedly)

		listValidated: list[int] = []

		for allegedInt in listInt_Allegedly:

			errorMessageContext: ErrorMessageContext = ErrorMessageContext(
				parameterValue = allegedInt,
				parameterValueType = type(allegedInt).__name__,
				isElement = True
			)

			# Always rejected as ambiguous
			if isinstance(allegedInt, bool):
				raise TypeError(errorMessageContext)

			# In this section, we know the Python type is not `int`, but maybe the value is clearly an integer.
			# Through a series of conversions, allow data to cascade down into either an `int` or a meaningful error message.

			if isinstance(allegedInt, (bytes, bytearray, memoryview)):
				errorMessageContext.parameterValue = None  # Don't expose potentially garbled binary data in error messages
				if isinstance(allegedInt, memoryview):
					allegedInt = allegedInt.tobytes()
				decodedString: CharsetMatch | None = charset_normalizer.from_bytes(allegedInt).best()
				if not decodedString:
					raise ValueError(errorMessageContext)
				allegedInt = errorMessageContext.parameterValue = str(decodedString)

			if isinstance(allegedInt, complex):
				if allegedInt.imag != 0:
					raise ValueError(errorMessageContext)
				allegedInt = float(allegedInt.real)
			elif isinstance(allegedInt, str):
				allegedInt = float(allegedInt.strip())

			if isinstance(allegedInt, float):
				if not float(allegedInt).is_integer():
					raise ValueError(errorMessageContext)
				allegedInt = int(allegedInt)
			else:
				allegedInt = int(allegedInt)

			listValidated.append(allegedInt)

			if lengthInitial is not None and isinstance(listInt_Allegedly, Sized) and len(listInt_Allegedly) != lengthInitial:
				raise RuntimeError((lengthInitial, len(listInt_Allegedly)))

	except (TypeError, ValueError) as ERRORmessage:
		if isinstance(ERRORmessage.args[0], ErrorMessageContext):
			context: ErrorMessageContext = ERRORmessage.args[0]
			if not context.containerType:
				context.containerType = type(listInt_Allegedly).__name__
			message = _constructErrorMessage(context, parameterName, parameterType)
			raise type(ERRORmessage)(message) from None
		# If it's not our Exception, don't molest it
		raise

	except RuntimeError as ERRORruntime:
		lengthCurrent: int
		lengthInitial, lengthCurrent = ERRORruntime.args[0]
		ERRORmessage = (
			f"The input sequence {parameterName} was modified during iteration. "
			f"Initial length {lengthInitial}, current length {lengthCurrent}."
		)
		raise RuntimeError(
			ERRORmessage
		) from None

	else:
		return listValidated

def oopsieKwargsie(huh: Any) -> bool | None | str:
	"""Interpret a `str` as `True`, `False`, or `None` to avoid an `Exception`.

	If a calling function passes a `str` to a parameter that shouldn't receive a `str`, `oopsieKwargsie()` might help you avoid an
	`Exception`. It tries to interpret the string as `True`, `False`, or `None`.

	Tests for this function can be run with: `from hunterMakesPy.tests.test_parseParameters import PytestFor_oopsieKwargsie`.

	Parameters
	----------
	huh : Any
		(huh) The input string to be parsed.

	Returns
	-------
	interpretedValue : bool | None | str
		The reserved keyword `True`, `False`, or `None` or the original string, `huh`.

	"""
	if not isinstance(huh, str):
		try:
			huh = str(huh)
		except BaseException:  # noqa: BLE001
			return huh
	formatted: str = huh.strip().title()
	if formatted == str(True):
		return True
	if formatted == str(False):
		return False
	if formatted == str(None):
		return None
	return huh



"""Configuration and fixtures for pytest.

(AI generated docstring)

This module provides shared fixtures and utility functions for the test suite,
including data paths, source code samples, and standardized assertion functions.

"""
# pyright: standard
from collections.abc import Callable
from typing import Any
import io
import pathlib
import pytest

# SSOT for test data paths and filenames
pathDataSamples: pathlib.Path = pathlib.Path("hunterMakesPy/tests/dataSamples")

# Fixture to provide a temporary directory for filesystem tests
@pytest.fixture
def pathTmpTesting(tmp_path: pathlib.Path) -> pathlib.Path:
	"""Provide a temporary directory for filesystem tests.

	(AI generated docstring)

	Parameters
	----------
	tmp_path : pathlib.Path
		The pytest built-in temporary path fixture.

	Returns
	-------
	pathTmpTesting : pathlib.Path
		The path to the temporary directory.

	"""
	return tmp_path

# Fixture for predictable Python source code samples
@pytest.fixture
def dictionaryPythonSourceSamples() -> dict[str, str]:
	"""Provide predictable Python source code samples for testing."""
	return {
		'functionFibonacci': "def fibonacciNumber():\n    return 13\n",
		'functionPrime': "def primeNumber():\n    return 17\n",
		'variablePrime': "prime = 19\n",
		'variableFibonacci': "fibonacci = 21\n",
		'classCardinal': "class CardinalDirection:\n    north = 'N'\n    south = 'S'\n",
	}

# Fixture for IO stream objects
@pytest.fixture
def streamMemoryString() -> io.StringIO:
	"""Provide a StringIO object for testing stream operations."""
	return io.StringIO()

# Fixture for predictable directory names using cardinal directions
@pytest.fixture
def listDirectoryNamesCardinal() -> list[str]:
	"""Provide predictable directory names using cardinal directions."""
	return ['north', 'south', 'east', 'west']

# Fixture for predictable file content using Fibonacci numbers
@pytest.fixture
def listFileContentsFibonacci() -> list[str]:
	"""Provide predictable file contents using Fibonacci sequence."""
	return ['fibonacci8', 'fibonacci13', 'fibonacci21', 'fibonacci34']

def uniformTestFailureMessage(expected: Any, actual: Any, functionName: str, *arguments: Any, **keywordArguments: Any) -> str:
	"""Format assertion message for any test comparison.

	Parameters
	----------
	expected : Any
		The expected value or outcome.
	actual : Any
		The actual value or outcome received.
	functionName : str
		The name of the function or test case being executed.
	*arguments : Any
		Positional arguments passed to the function having its return value checked.
	**keywordArguments : Any
		Keyword arguments passed to the function having its return value checked.

	Returns
	-------
	message : str
		A formatted failure message detailing the expectation vs reality.

	"""
	listArgumentComponents: list[str] = [str(parameter) for parameter in arguments]
	listKeywordComponents: list[str] = [f"{key}={value}" for key, value in keywordArguments.items()]
	joinedArguments: str = ', '.join(listArgumentComponents + listKeywordComponents)

	return (f"\nTesting: `{functionName}({joinedArguments})`\n"
			f"Expected: {expected}\n"
			f"Got: {actual}")

def standardizedEqualTo(expected: Any, functionTarget: Callable[..., Any], *arguments: Any, **keywordArguments: Any) -> None:
	"""Template for most tests to compare actual outcome with expected outcome.

	Includes handling for expected errors/exceptions.

	Parameters
	----------
	expected : Any
		The expected return value, or an Exception type if an error is expected.
	functionTarget : Callable[..., Any]
		The function to call and test.
	*arguments : Any
		Positional arguments to pass to `functionTarget`.
	**keywordArguments : Any
		Keyword arguments to pass to `functionTarget`.

	"""
	if type(expected) == type[Exception]:  # noqa: E721
		messageExpected: str = expected.__name__
	else:
		messageExpected = expected

	try:
		messageActual = actual = functionTarget(*arguments, **keywordArguments)
	except Exception as actualError:
		messageActual = type(actualError).__name__
		actual = type(actualError)

	functionName: str = getattr(functionTarget, "__name__", functionTarget.__class__.__name__)
	assert actual == expected, uniformTestFailureMessage(messageExpected, messageActual, functionName, *arguments, **keywordArguments)

# Why I wish I could figure out how to implement standardized* test functions.
# ruff: noqa: ERA001
	# standardizedEqualTo(expected, updateExtendPolishDictionaryLists, *value_dictionaryLists, **keywordArguments)
# NOTE one line of code with `standardizedEqualTo` (above) replaced the following ten lines of code. Use `standardizedEqualTo`.
	# if isinstance(expected, type) and issubclass(expected, Exception):
	#	 with pytest.raises(expected):
	#		 updateExtendPolishDictionaryLists(*value_dictionaryLists, **keywordArguments)
	# else:
	#	 result = updateExtendPolishDictionaryLists(*value_dictionaryLists, **keywordArguments)
	#	 if description == "Set values":  # Special handling for unordered sets
	#		 for key in result:
	#			 assert sorted(result[key]) == sorted(expected[key])
	#	 else:
	#		 assert result == expected
